#include "file.h"
#include <ctype.h>

//constructors
File::File(void)
{
    perror("no filename or mode given");
}

void File::expand_tabs(char *input_buffer)
{
    int length = strlen(input_buffer);
    int number_of_tabs = 0;

    for (int i=0;i<length;i++)
    {
        if (input_buffer[i] == TAB)
            number_of_tabs++;
    }
    if ((length + (TABSIZE -1) *number_of_tabs) > (INPUT_BUFFERSIZE-1))
        return; // don't do it, it won't fit
    
    
    char *copy = new char[INPUT_BUFFERSIZE];
    strcpy(copy, input_buffer);
    int count = 0;
    for (int i=0; i<=length; i++)
    {
        if (copy[i]==TAB)
        {
            for (int j=0;j<TABSIZE;j++)
            {
                input_buffer[count+j] = ' ';
            }
            count += TABSIZE;
        }
        else
        {
            input_buffer[count] = copy[i];
            count++;
        }
    }
    delete [] copy;
}



File::File(char *filename, int _mode,char *_commentchar)
{
    FILE *infile;
    char *input_buffer = new char[INPUT_BUFFERSIZE];
    comment = strdup(_commentchar);
    Line = NULL;
    number_of_lines = 0;
    mode = _mode;
    search_length = 0;
    char message_buffer[256];
    if (!(infile = fopen(filename, "r")))
    {
        perror("error opening file");
    }

    Line = (char **)realloc(Line,(number_of_lines+1) * sizeof (char *));
    while(fgets (input_buffer, INPUT_BUFFERSIZE, infile))
    {
        if (EXPAND_TABS)
            expand_tabs(input_buffer);
        if ((mode & MODE_CODE) || (mode & MODE_REMOVE_ALL_SPACES))
            remove_spaces(input_buffer);
        //remove all CR/LF's and ascii 7's (beep) 
        for (int i=0;input_buffer[i];i++)
        {
            if (input_buffer[i] == 10 || input_buffer[i] == 13 || input_buffer[i] == 7)
                input_buffer[i] = ' ';
        }
        number_of_lines++;
        /*******************************************************************
        mooier zou zijn om de file eerst in een doublelinked-list op te slaan
        en daarna(als het aantal lines bekend is) een array van pointers naar 
        lines op te bouwen.
        (hetzelfde geldt ook voor input_buffer)
        wegens tijdgebrek heb ik het maar even zo gedaan
        *******************************************************************/
        Line = (char **)realloc(Line,(number_of_lines+1) * sizeof (char *));
        Line[number_of_lines - 1] = strdup(input_buffer);
        if (! (number_of_lines % 1000))
        {
            sprintf(message_buffer," Read %d lines.",number_of_lines);
            message(message_buffer);
        }
    }
    delete [] input_buffer;
    fclose (infile);
    Line[number_of_lines] ="";

    line_tag = new int[number_of_lines + 1];
    for (int i=0; i <= number_of_lines; i++)
    {
        line_tag[i] = -1;
    }

}

File::File(File const &other)
{
    warning("copying LARGE File object"); 
    copy(other);
}

File const &File::operator=(File const &other)
{
    if (this != &other)
    {
        destroy();
        copy(other);
    }
    return (*this);
}

File::~File()
{
    destroy();
}

void File::copy(File const &other)
{
    Line = (char **)malloc(sizeof(char *) * other.get_number_of_lines());
    for (int i=0; i <= other.get_number_of_lines(); i++) 
    {
        Line[i] = strdup(other.line(i));
    }
    line_tag = new int[other.get_number_of_lines() + 1];
    for (int i=0; i <= get_number_of_lines(); i++)
    {
        line_tag[i] = other.get_tag(i);
    }
}

void File::destroy()
{
    for (int i=0; i<=number_of_lines; i++)
        free (Line[i]);
    free (Line);
    delete [] line_tag;
    delete comment;
}

char *File::remove_spaces(char *input_buffer)
{
    char *string = input_buffer;
    char *tmp;
    //skip leading spaces(they'not removed)
    while(isspace(*string))
        string++;
    //then remove internal spaces +CR/LF
    do
    {
        while (isspace(*string) && isspace(*(string+1)))
            for (tmp = string; *tmp; tmp++)
                *tmp = *(tmp + 1);
        while ((*string) == 13 || (*string) == 10)
            for (tmp = string; *tmp; tmp++)
                *tmp = *(tmp + 1);
    }
    while(*(string++));
    return input_buffer;
}

int File::get_number_of_lines() const
{
    return number_of_lines;
}

int File::get_tag(int number) const
{
    return line_tag[number];
}

//compare will return false for any string shorter or equal than num characters 


int File::compare(int number, const char *string) const
{
    if (number > number_of_lines) 
        number = number_of_lines;
    char 
        *tmp_string = strdup(Line[number]);
    char *tmp_string_backup_for_removal_at_end_of_funtion = tmp_string;


    if ((mode & MODE_CODE) || (mode & MODE_REMOVE_ALL_SPACES))
    {
        //cut off all leading spaces
        while(isspace(*tmp_string))
            tmp_string++;
    }
    if (mode & MODE_REMOVE_ALL_SPACES)
    {
        //remove all spaces
        // this loop stops at the end-zero of the string
        for (char *tmp_string_copy = tmp_string;*tmp_string_copy;tmp_string_copy++)
        {
            if (isspace(*tmp_string_copy))
            {
                for (int count = 0;tmp_string[count];count++)
                {
                    tmp_string_copy[count] = tmp_string_copy[count+1]; 
                }
            }
        }
    }
    if (mode & MODE_IGNORE_COMMENT)
        remove_eol_comment(tmp_string);

    if (strlen(string) != strlen(tmp_string))
    {
        delete tmp_string_backup_for_removal_at_end_of_funtion;
        return 0;
    }
    int retvalue = !strcmp (string, tmp_string);
    delete tmp_string_backup_for_removal_at_end_of_funtion;
    return retvalue;

}

/****************************************************************************
 * deze functie zoekt naar Line[x] die matcht met string
 * en waarvoor x zo dicht mogelijk bij number ligt
 * regels waarvan het line_tag[x] niet -1 is worden niet meegenomen
 * (die zijn al gematcht)
 ***************************************************************************/
int File::search(int number, const char *string) const
{
    int 
        count_up, 
        count_down;
    count_up = count_down = number;
    char *tmp_string = strdup(string);
    char *tmp_string_backup_for_removal_at_end_of_function = tmp_string;
    if ((mode & MODE_CODE) || (mode & MODE_REMOVE_ALL_SPACES))
    {
        //cut off all leading spaces
        while(isspace(*tmp_string))
            tmp_string++;
    }

    if (mode & MODE_REMOVE_ALL_SPACES)
    {
        //remove all spaces
        // this loop stops at the end-zero of the string
        for (char *tmp_string_copy = tmp_string;*tmp_string_copy;tmp_string_copy++)
        {
            if (isspace(*tmp_string_copy))
            {
                for (int scount = 0;tmp_string[scount];scount++)
                {
                    tmp_string_copy[scount] = tmp_string_copy[scount+1]; 
                }
            }
        }
    }
    if (mode & MODE_IGNORE_COMMENT)
        remove_eol_comment(tmp_string);
    if ((int)strlen(tmp_string) <= search_length)  //only search in this or the next line
    {
        if ((line_tag[number] < 0) && (compare(number,tmp_string)))
        {
            delete tmp_string_backup_for_removal_at_end_of_function;
            return number;
        }
        if ((line_tag[number + 1] < 0) && (compare(number + 1,tmp_string)))
        {
            delete tmp_string_backup_for_removal_at_end_of_function;
            return number + 1;
        }
        delete tmp_string_backup_for_removal_at_end_of_function;
        return - 1;
    }
    while(1)
    {
        if (count_down >= 0)
        {
            if (line_tag[count_down] < 0)
            {
                if (compare(count_down, tmp_string))
                {
                    delete tmp_string_backup_for_removal_at_end_of_function;
                    return count_down;
                }
            }
            count_down--;
        }
        if (count_up < number_of_lines)
        {
            if (line_tag[count_up] < 0)
            {
                if (compare(count_up, tmp_string))
                {
                    delete tmp_string_backup_for_removal_at_end_of_function;
                    return count_up;
                }
            }
            count_up++;
        }
        if ((count_up >= number_of_lines) && (count_down < 0))
        {
            delete tmp_string_backup_for_removal_at_end_of_function;
            return -1; //string not found
        }
    }
    delete tmp_string_backup_for_removal_at_end_of_function; //don't think this is needed, but it doesn't hurt to put it here(will be optimized out anyway)
}

void File::set_tag(int _line, int _tag)
{
    if (_line > get_number_of_lines()) 
        _line = get_number_of_lines();
    line_tag[_line] = _tag;
}


const char *File::line(int number) const
{
    if (number >number_of_lines) number = number_of_lines;
    return Line[number];
}

void File::s_length(int _l)
{
    search_length = _l;
}

void File::remove_eol_comment(char *string) const// this function removes //comment  unless it starts on the first char of the line
{
    for (char *tmp_string_copy = string + 1;*tmp_string_copy;tmp_string_copy++)
    {
        if (*tmp_string_copy == comment[0])
        {
            if (!strncmp(tmp_string_copy,comment,strlen(comment)))
            {
                *tmp_string_copy = 0;
                return;
            }
        }
    }
}


void File::insert_line(int number,char const *line)
{
        Line = (char **)realloc(Line,(number_of_lines+2) * sizeof (char *));
        line_tag = (int *)realloc(line_tag,(number_of_lines+2) * sizeof (int));
        for (int i=number_of_lines+1;i>number;i--)
        {
            Line[i] = Line[i - 1];
            line_tag[i] = line_tag[i - 1];    
        }
        Line[number] = strdup(line);
        line_tag[number] = -1;
        number_of_lines++;
}

void File::delete_line(int number)
{
        if (number == number_of_lines)
            return;
        number_of_lines--;
        delete Line[number];
        for (int i=number;i<=number_of_lines;i++)
        {
            Line[i] = Line[i + 1];
            line_tag[i] = line_tag[i + 1];    
        }
        Line = (char **)realloc(Line,(number_of_lines+1) * sizeof (char *));
        line_tag = (int *)realloc(line_tag,(number_of_lines+1) * sizeof (int));

}
