/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#ifndef _SECRETARY_h_
#define _SECRETARY_h_
#include "wobject.h"
#include <scrsav.h>

class Scheduler_List
{
    public:
        friend class Secretary;
        Scheduler_List(int (*f)(void *arg),void *arg,int _time_increment,Scheduler_List *head);
        ~Scheduler_List();
    private:
        void do_list();
        int (*function)(void *arg);
        void *arg;
        int next_execute_time;
        int time_increment;
        Scheduler_List *last;
        Scheduler_List *next;
};


class Secretary
{
    public:
        friend class Scheduler_List;
        friend void ScreenStimer(...);
        Secretary();
        ~Secretary();
        Wobject *process_mouse(int *x = NULL, int *y = NULL);
        static void set_key_catching(int catch_keys);
        static void set_screen_saver(void (*)(),int time, char *name = NULL);
        static int load_screen_saver(char *filename, int time);
        static void reset_screen_saver_time(int time);
        static void set_screensaver_suspension(int susp);
        static void schedule(int (*fun)(void *arg),void *arg,int time);
        static void destroy();
        static void create();
        static char *get_current_screen_saver_name();
        static int initialized;
    private:
        static Wobject *mouse_object;
        static int init_timer;
        static Scheduler_List *schedulerlist;
        static int ScreenScounter;
        static int ScreenStime;
        static int SecCounter;
        static void (*screensaver)();
        static int catch_keys;
        static int old_mouse_x,old_mouse_y;
        static int screensaver_suspended;
        static char **loaded_screensaver_filenames;
        static char **loaded_screensaver_names;
        static void **loaded_screensavers;
        static int number_of_loaded_screensavers;
        static char *current_screen_saver_name;
};

extern Secretary Wham_Secretary;

#endif
