/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#ifndef _window_h_
#define _window_h_
#include "helpers.h"
#include "wobject.h"

#define WINDOW_SCROLL_SHIFT 1
#define WINDOW_SCROLL_OVERLAY 0


class Window : public Wobject
{
    public:
        Window(char *name,int size_x,int size_y,int pos_x,int pos_y);
        Window(char *name,int size_x,int size_y,int pos_x,int pos_y, WinSystem *its_manager);
        Window(char *name,BITMAP *bitmap,int pos_x,int pos_y); // size of bitmap is used for the size of the window
        Window(char *name,BITMAP *bitmap, int _pos_x, int _pos_y, WinSystem *_its_manager);

        Window(char *name,int size_x,int size_y,int pos_x,int pos_y,int (*draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y));
        Window(char *name,int size_x,int size_y,int pos_x,int pos_y, WinSystem *its_manager,int (*draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y));
        Window(char *name,BITMAP *bitmap,int pos_x,int pos_y,int (*draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)); // size of bitmap is used for the size of the window
        Window(char *name,BITMAP *bitmap, int _pos_x, int _pos_y, WinSystem *_its_manager,int (*draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y));
        ~Window();

        // get variouys objects        
        BITMAP *bmp();     // returns the user area of the object

        //user area 
        void set_user_area(int pos_x, int pos_y, int size_x, int size_y);

        // resizing/rescaling
        int resize(int nw_size_x, int nw_size_y);


        // extra functions
        int printf(char *format,...); // prints text on the current cursor position

        //statics
        static int Text_Color;
        static int Scroll_Behaviour;

        // publics
        int background_color;

        // process input/messages
        Wobject *process_mouse(int *x_out = NULL, int *y_out = NULL);
    private:
        void draw(BITMAP *onto, int x_from, int y_from, int x_to, int y_to); // draws (a part of) the object the coordinates are screen coordinates
        

        BITMAP *window;                 // buffer
        BITMAP *user;                   // user area(sub bitmap of window)

        int must_delete_bitmap;
        void create(BITMAP *bmp); // creates a object with bitmap as buffer
        void destroy();            // just does what it says
        int text_cursor_x,text_cursor_y;
};



#endif
