/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/


#ifndef _winsys_h_
#define _winsys_h_
#include <stdlib.h>
#include <allegro.h>
#define WHAM_NO_SHADOW     0
#define WHAM_SIMPLE_SHADOW -1  
#define WHAM_FAST_SHADOW   1
#define WHAM_SLOW_SHADOW   2
#define WHAM_PATTERN_SHADOW   -2


#define W_MSG_RAISE             -1
#define W_MSG_HIDE              -2
#define W_MSG_SHOW              -3
#define W_MSG_UPDATE            -4
#define W_MSG_RAISED            -5
#define W_MSG_HIDDEN            -6
#define W_MSG_SHOWN             -7
#define W_MSG_MOVED             -8
#define W_MSG_MOVE_RESCALED     -9
#define W_MSG_RESIZED           -10  
#define W_MSG_RESCALED          -11
#define W_MSG_VIEWPORT_RESIZED  -12
#define W_MSG_CREATED           -13
#define W_MSG_DESTROYED         -14
#define W_MSG_USER_AREA_SET     -15
#define W_MSG_LOWERED           -16
#define W_MSG_LOWER             -17
#define W_MSG_ATTACHED          -18
#define W_MSG_SHIFTED           -19
#define W_MSG_MANAGER_DELETE    -20
#define W_MSG_VIEWPORT_MOVED    -21
#define W_MSG_GOING_TO_RAISE    -22
#define W_MSG_GOING_TO_LOWER    -23
#define W_MSG_UPDATED           -24
#define W_MSG_LOCKED            -25
#define W_MSG_UNLOCKED          -26
#define W_MSG_RAISE_AND_LOCKED  -27
#define W_MSG_LOSTMOUSE         -28
#define W_MSG_GOTMOUSE          -29
#define W_MSG_FROZEN            -30
#define W_MSG_THAWN             -31

// message added?? don't forget to update helper message_string(   ) function

//display modes
#define DISPLAY_MODE_FULL 1
#define DISPLAY_MODE_NOBORDER 2

#include "wobject.h"
#include "topinc.h"
#include "w_mousec.h"

// effects for show/hide

#define W_EFFECT_NO_EFFECT 0
#define W_EFFECT_HLINES  1
#define W_EFFECT_VLINES  2
#define W_EFFECT_DOTS    3
#define W_EFFECT_PRERECT 4
#define W_EFFECT_BUNCH   5
extern int wham_trace;

class WinSystem
{
    public:
        friend class Wobject;
        friend class Boxlist;
        friend void mouse_on();
        

        // constructor/destructor
        WinSystem();
        WinSystem(Wobject *parent);
        ~WinSystem();

        // these statics must be set ONCE before creating the first WinSystem
        static int UseBoxes;
        static int UseSpecialSegvHandler;


        // these statics must be set to the desired values before the creation
        // of *each* WinSystem (really they are arguments of the constructor,
        // but put here to avoid extremely long constructors)
        static int InitialBufferSize;  // the inial size of the adminitration tables of this class
        static int Buffered;            // if on, do all operations on a buffer and blit the result to TheScreen
        static int ShadowStyle;        // determines the type of shadowing used
        static int ShadowColor;        // determines the type of shadowing used
        static int ShadowSize;         // determines the type of shadowing used
        static int BackgroundColor;
        static BITMAP *TheScreen;
        
        
        // these settings may be changed during the program
        // and will take effect immediately for all WinSys objects
        static int Buffer_Increment;    // the amount the buffers will grow when exceeded
        static int Full_Remove;        // only set to false when using root windows
        static int Minimum_Size_x;      // minimum sizes, must be > 0
        static int Minimum_Size_y;        
        static int Message_System_Active;        
        static int Show_Effect;
        static int Show_Effect_Param;
        static int Vsynced;
        static Mouse_Cursor              // can be used to change mouse pointer
            *Mouse_Cursor0,        // these are initially set to NULL
            *Mouse_Cursor1,        // but can be redefined in main program
            *Mouse_Cursor2,        // suggestion is to use Mouse_Cursor 1 for
            *Mouse_Cursor3,        // resizing actions and borderwin 2 for moving
            *Mouse_Cursor4,        // action (the BorderWin class does this)
            *Mouse_Cursor5,        // all my mouse handlers set the mouse
            *Mouse_Cursor6,        // to Mouse_Cursor0 when it is not inside a
            *Mouse_Cursor7,        // window as long as this is NULL this is the
            *Mouse_Cursor8,        // default arrow, but you can set it to something
            *Mouse_Cursor9;        // else
        
        
        //user functions
        // info functions
        static char *get_version();
        int get_TheScreen_size_x();
        int get_TheScreen_size_y();
        int get_number_of_wobjects();  // return number of object this manager has
        // functions for getting various objects
        BITMAP *get_TheScreen(); 
        Wobject *get_wobject(int tag);
        Wobject *get_parent();
        Wobject *get_top_wobject();   // return top object not including top-locked objs
        Wobject *get_wobject_from_hierarchy(int hierarchy); // returns object in a special place in the drawing hierachy
        Wobject *get_keyboard_claimer();

        // coordinate translator function
        Wobject *process_coordinates(int x, int y, int *out_x, int *out_y, int *flag);

        // actions concerning all objects
        void refresh_all_visible_objects(); // refreshes all visible objects
        void shift_all(int x, int y);
        
        // message functions
        int send_message(Wobject *from,int MSG,Wobject *to, Wobject **responded = NULL);  // to=NULL means to all
        int broadcast_message(Wobject *from,int MSG);
        void set_user_msg_processor(int (*user_process_message)(Wobject *from, int MSG));
       
        // gfx mode switches
        void switch_gfx_mode(int mode,int x, int y,int colordepth = 0);
        void let_user_choose_gfx_mode();

        //helper function
        BITMAP *get_scratch_bmp(int size_x,int  size_y, int color_depth = 0);
        
        // low level functions (USE WITH CARE)
        void remove_wobject(int tag);  // removes the object from the screen
        void remove_wobject(int tag,int xfrom,int yfrom,int xto,int yto);  // removes part of the object from the screen
        void display_wobject(int tag,int mode);  //draws the object to the screen mode = DISPLAY_MODE_FULL or DISPLAY_MODE_NOBORDER  (updates the whole window , or just without the border)
        void display_wobject(int tag,int xfrom,int yfrom,int xto,int yto);
        void update_screen(int xf,int yf,int xt,int yt); //for use when buffered, blits part of buffer to screen
        void update_screen(int xf,int yf,int xt,int yt,int effect); //for use when buffered, blits part of buffer to screen
		  void claim_coordinates(Wobject *which); // mouse will only see this window
		  void release_coordinates();             // mouse will see all windows again
        void claim_keyboard(Wobject *which); // all keypresses will go to this object
        void release_keyboard();
        



        // for debugging purposes
        void show_all_boxlist_boxes();
        void defrag_all_boxlist_boxes();
    
    
    
        static int Turn_Off_Mouse;      // OBSOLETE only here for compatability reasons
    private:  // only the wobject class has access to some of these functions

        static Wobject *coordinate_claimer;
        static Wobject *keyboard_claimer;
        static BITMAP *scratch_bmp;
        static char *version;
        static int must_delete_colormap;
        int x_shifted,y_shifted;  // administration of shift_all function
        int get_new_tag(Wobject *_wobject);
        int is_being_deleted; // to prevent circular deletion from wobject
        void destroy_tag(int tag);
        void raise(int tag);
        void lower(int tag);
        void raise_guaranteed(int tag);
        void rebuild_screen(BITMAP *onto,int xfrom,int yfrom,int xto,int yto, int from_wobject);
        void rebuild_screen_using_boxes(BITMAP *onto,int xfrom,int yfrom,int xto,int yto, int from_wobject);
        int get_order(int tag); // returns the place in the drawing hiearchy
        // info on this instance of the class
        int _InitialBufferSize;  // the inial size of the adminitration tables of this class
        int _Buffered;            // if on, do all operations on a buffer and blit the result to TheScreen
        int _ShadowStyle;        // determines the type of shadowing used
        int _ShadowColor;        // determines the type of shadowing used
        int _ShadowSize;         // determines the type of shadowing used
        int _BackgroundColor;        

        BITMAP *this_managers_screen;
        BITMAP *shadow_pattern;
        
        // very private, even wobject shouldn't touch these
        Wobject *parent;        
        
        Wobject **wobject_list;    // list of all wobjects, tag is index in this list
        int *wobject_order_list; // list of all wobjects, tag of wobject is index in this array, value of each array position lists it's place in the order list
        int *order_wobject_list; // list of wobject order, index 0 is top wobject value is tag
        int number_of_wobjects;
        int buffersize;
        static int winsys_count;
        int get_free_tag();
        int get_number_of_locked_top_wobjects();
        int get_first_free_bottom_place();
        BITMAP *buffer;
        int TheScreen_size_x;
        int TheScreen_size_y;
        void create();
        void parent_bitmap_has_changed();
        static int (*user_process_message)(Wobject *from, int MSG);


        //boxlist stuff
        void update_boxlists(int from);
        void update_boxlists(Wobject *from);
        void redraw_higher_lying_see_through_objects(BITMAP *onto,int fomx,int fromy,int tox, int toy,int  from);
        void redraw_higher_lying_see_through_objects(BITMAP *onto,int fomx,int fromy,int tox, int toy,Wobject *from);
        void redraw_higher_lying_objects(BITMAP *onto,int fomx,int fromy,int tox, int toy,int  from);
        void redraw_higher_lying_objects(BITMAP *onto,int fomx,int fromy,int tox, int toy,Wobject *from);

};




#endif
