/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/


#include <stdio.h>
#include <strings.h>
#include "../include/Bordered.h"
#include "../include/whamdat.h"

// some variables from allegro, which are not officially exported
extern BITMAP *_mouse_screen;
extern int _color_conv;

int BorderWin::count = -1;
DATAFILE *BorderWin::datafile = NULL;
BITMAP *BorderWin::upper_border = NULL;
BITMAP *BorderWin::lower_border = NULL;
BITMAP *BorderWin::left_border = NULL;
BITMAP *BorderWin::right_border = NULL;
int BorderWin::upper_border_size;
int BorderWin::lower_border_size;
int BorderWin::left_border_size;
int BorderWin::right_border_size;
SystemBox **BorderWin::systemboxes = NULL;
int  BorderWin::number_of_system_boxes = 0;

int BorderWin::RemapPalette = TRUE;
char *BorderWin::DatafileName = NULL;
BITMAP *BorderWin::UpperBorderBitmap = NULL;
BITMAP *BorderWin::LowerBorderBitmap = NULL;
BITMAP *BorderWin::RightBorderBitmap = NULL;
BITMAP *BorderWin::LeftBorderBitmap = NULL;
BITMAP *BorderWin::UpperLeftCorner = NULL;
BITMAP *BorderWin::UpperRightCorner = NULL;
BITMAP *BorderWin::LowerLeftCorner = NULL;
BITMAP *BorderWin::LowerRightCorner = NULL;
int BorderWin::max_ver_borderlength = 0;
int BorderWin::max_hor_borderlength = 0;
int BorderWin::Move_Style = BORDERWIN_OPAQUE;


BorderWin::BorderWin(int _size_x,int _size_y,int _pos_x, int _pos_y, char *_name)
    : Window(_name,_size_x,_size_y,_pos_x,_pos_y , BorderWin::draw_border)
{
    create();
}

BorderWin::BorderWin(int _size_x,int _size_y,int _pos_x, int _pos_y, char *_name, WinSystem *its_manager)
    : Window(_name,_size_x,_size_y ,_pos_x,_pos_y,its_manager, BorderWin::draw_border)
{
    create();
}


BorderWin::BorderWin(BITMAP *w,int _pos_x, int _pos_y, char *_name)
    : Window(_name,w,_pos_x,_pos_y, BorderWin::draw_border)
{
    create();
}

BorderWin::BorderWin(BITMAP *w,int _pos_x, int _pos_y, char *_name, WinSystem *its_manager)
    : Window(_name,w,_pos_x,_pos_y,its_manager, BorderWin::draw_border)
{
    create();
}

void BorderWin::create()
{
    count++;
    if (!systemboxes)
    {
        systemboxes = (SystemBox **)malloc(sizeof(SystemBox *) * (1));
        if (!systemboxes)
            out_of_mem("cannot create system boxes");
        systemboxes[0] = NULL;

    }
}


BorderWin::~BorderWin()
{
    if (!count)
    {
        if (datafile)
            unload_datafile(datafile);
        else
            destroy_bitmap(UpperBorderBitmap); // delete bogus bitmap

        destroy_system_boxes();
        if (its_palette) free(its_palette);
        if (upper_border) destroy_bitmap(upper_border);
        if (lower_border) destroy_bitmap(lower_border);
        if (right_border) destroy_bitmap(right_border);
        if (left_border) destroy_bitmap(left_border);
        if (DatafileName)
            delete DatafileName;
    }
    count--;
}

void BorderWin::add_system_box(BITMAP *bmp,void (_function)(Wobject *win))
{
    if (!systemboxes)
    {
        number_of_system_boxes = 0;
        systemboxes = (SystemBox **)malloc(sizeof(SystemBox *) * (number_of_system_boxes +1));
        if (!systemboxes)
            out_of_mem("cannot add systembox");
        systemboxes[0] = NULL;
    }
    number_of_system_boxes++;
    systemboxes = (SystemBox **)realloc(systemboxes, sizeof(SystemBox *) * (number_of_system_boxes +1));
    if (!systemboxes)
        out_of_mem("cannot add systembox");
    systemboxes[number_of_system_boxes] = NULL;
    systemboxes[number_of_system_boxes - 1] = new SystemBox(bmp,_function);
    if (!systemboxes[number_of_system_boxes - 1])
        out_of_mem("cannot add systembox");

}

void BorderWin::destroy_system_boxes()
{
    if (systemboxes)
    {
        for (int i=0;systemboxes[i];i++)
        {
            delete systemboxes[i];
        }
        free (systemboxes);
        systemboxes = NULL;
        number_of_system_boxes = 0;
    }
}

int BorderWin::draw_border(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)
{
    int colconv = _color_conv;
    GeoInfo const *g = the_object->get_geoinfo();
    if (!UpperBorderBitmap)
    {
        _color_conv = COLORCONV_NONE;
        datafile = DatafileName ? load_datafile(DatafileName) : load_datafile("wahmdat.dat");
        _color_conv = colconv;
        if (!datafile)
        {
            RGB *f,*b,*n;            
            f = new RGB;                                    
            b = new RGB;                                    
            n = new RGB;                                    
            get_color(gui_fg_color,f);
            get_color(gui_bg_color,b);
            n->r = n->b = n->g = 63;
            set_color(gui_bg_color,n);
            n->r = n->b = n->g = 0;
            set_color(gui_fg_color,n);
            alert("Unable to open",DatafileName, "using bogus borders.","OK",NULL,0,0);
            set_color(gui_bg_color,b);
            set_color(gui_fg_color,f);

            UpperBorderBitmap =
            LowerBorderBitmap =
            LeftBorderBitmap = 
            RightBorderBitmap = 
            UpperLeftCorner =
            UpperRightCorner = 
            LowerLeftCorner = 
            LowerRightCorner =
            create_bitmap(10,10);
            if 
            (
                !UpperBorderBitmap ||
                !LowerBorderBitmap ||
                !LeftBorderBitmap ||
                !RightBorderBitmap ||
                !UpperLeftCorner ||
                !UpperRightCorner ||
                !LowerLeftCorner ||
                !LowerRightCorner
            )
                out_of_mem("creating BorderWin borders");
            delete f;
            delete n;
            delete b;
        }
        else
        {
            UpperBorderBitmap = ((BITMAP *)(datafile[UPPER_BORDER_BITMAP].dat));
            LowerBorderBitmap = ((BITMAP *)(datafile[LOWER_BORDER_BITMAP].dat));
            LeftBorderBitmap = ((BITMAP *)(datafile[LEFT_BORDER_BITMAP].dat));
            RightBorderBitmap = ((BITMAP *)(datafile[RIGHT_BORDER_BITMAP].dat));
            UpperLeftCorner = ((BITMAP *)(datafile[UPPER_LEFT_CORNER].dat));
            UpperRightCorner = ((BITMAP *)(datafile[UPPER_RIGHT_CORNER].dat));
            LowerLeftCorner = ((BITMAP *)(datafile[LOWER_LEFT_CORNER].dat));
            LowerRightCorner = ((BITMAP *)(datafile[LOWER_RIGHT_CORNER].dat));
            if 
            (
                !UpperBorderBitmap ||
                !LowerBorderBitmap ||
                !LeftBorderBitmap ||
                !RightBorderBitmap ||
                !UpperLeftCorner ||
                !UpperRightCorner ||
                !LowerLeftCorner ||
                !LowerRightCorner
            )
                out_of_mem("creating BorderWin borders");

            if (RemapPalette)
            {
                remap_bitmap(UpperBorderBitmap,(PALETTE)(datafile[BORDER_PALETTE].dat));
                remap_bitmap(LowerBorderBitmap,(PALETTE)(datafile[BORDER_PALETTE].dat));
                remap_bitmap(RightBorderBitmap,(PALETTE)(datafile[BORDER_PALETTE].dat));
                remap_bitmap(LeftBorderBitmap,(PALETTE)(datafile[BORDER_PALETTE].dat));
                remap_bitmap(UpperLeftCorner,(PALETTE)(datafile[BORDER_PALETTE].dat));
                remap_bitmap(UpperRightCorner,(PALETTE)(datafile[BORDER_PALETTE].dat));
                remap_bitmap(LowerLeftCorner,(PALETTE)(datafile[BORDER_PALETTE].dat));
                remap_bitmap(LowerRightCorner,(PALETTE)(datafile[BORDER_PALETTE].dat));
                its_palette = (RGB *)malloc(256*sizeof(RGB));
                if (!its_palette)
                    out_of_mem("cannot create palette");
                get_palette(its_palette);
            }
            else
                its_palette = (RGB *)(datafile[BORDER_PALETTE].dat);
            

        } //end if file open succeded
    }// end init border bitmaps


    if (!onto)
    {
        upper_border_size = UpperBorderBitmap->h;
        lower_border_size = LowerBorderBitmap->h;
        left_border_size = LeftBorderBitmap->w;
        right_border_size = RightBorderBitmap->w;
        switch(from_x)
        {
            case WINDOW_UPPER_BORDER:
                return upper_border_size;
            break;
            case WINDOW_LOWER_BORDER:
                return lower_border_size;
            break;
            case WINDOW_LEFT_BORDER:
                return left_border_size;
            break;
            case WINDOW_RIGHT_BORDER:
                return right_border_size;
            break;
        }
    }
    if (g->screen_size_x > max_hor_borderlength || g->screen_size_y>max_ver_borderlength)
    {
        max_hor_borderlength = max_hor_borderlength > g->screen_size_x ? max_hor_borderlength : g->screen_size_x;
        max_ver_borderlength = max_ver_borderlength > g->screen_size_y ? max_ver_borderlength : g->screen_size_y;
        if (upper_border)
        {
            destroy_bitmap(upper_border);
            destroy_bitmap(lower_border);
            destroy_bitmap(left_border);
            destroy_bitmap(right_border);
        }
        upper_border = create_bitmap(max_hor_borderlength, upper_border_size);
        left_border = create_bitmap(left_border_size,max_ver_borderlength);
        lower_border = create_bitmap(max_hor_borderlength , lower_border_size);
        right_border = create_bitmap(right_border_size,max_ver_borderlength);

        if (!upper_border || !lower_border || !left_border || !right_border)
            out_of_mem("buffers for border drawing");
        clear(upper_border);
        clear(lower_border);
        clear(left_border);
        clear(right_border);

        if (its_palette && bitmap_color_depth(onto) !=8)
        {
                select_palette(its_palette);
        }
        for (int i=0;i<max_hor_borderlength;i+=LowerBorderBitmap->w)
        {
            draw_sprite(lower_border,LowerBorderBitmap,i,0);
        }
        for (int i=0;i<max_ver_borderlength;i+=LeftBorderBitmap->h)
        {
            draw_sprite(left_border,LeftBorderBitmap,0,i);
        }
        for (int i=0;i<max_ver_borderlength;i+=RightBorderBitmap->h)
        {
            draw_sprite(right_border,RightBorderBitmap,0,i);
        }
        if (its_palette && bitmap_color_depth(onto) !=8)
        {
                unselect_palette(its_palette);
        }
    }
    // upper border needs redrawing each time it is used because of title and system boxes

    if (its_palette && bitmap_color_depth(onto) !=8 )
    {
        select_palette(its_palette);
    }
    clear(upper_border);
    for (int i=0;i<max_hor_borderlength;i+=UpperBorderBitmap->w)
    {
        draw_sprite(upper_border,UpperBorderBitmap,i,0);
    }

    if (g->name)
    {
        text_mode(-1);
        textout(upper_border,font,g->name,2,1,makecol(255,255,255));
    }
    int s_pos = g->screen_size_x;
    for (int i=0;systemboxes[i];i++)
    {
        s_pos -= systemboxes[i]->sprite->w;
        draw_sprite(upper_border,systemboxes[i]->sprite,s_pos,0);
    }
    int bfrom_x,bfrom_y,bto_x,bto_y;
    // upper border
    bfrom_x = MAX(g->pos_x,from_x);
    bfrom_y = MAX(g->pos_y - upper_border_size, from_y);
    bto_x = MIN(g->pos_x + g->screen_size_x - 1,to_x);
    bto_y = MIN(g->pos_y - 1,to_y);
    blit(upper_border,onto,bfrom_x - g->pos_x , bfrom_y - g->pos_y + upper_border_size, bfrom_x,bfrom_y, bto_x - bfrom_x + 1, bto_y - bfrom_y + 1); 

    // lower border
    bfrom_x = MAX(from_x,g->pos_x);
    bfrom_y = MAX(from_y,g->pos_y + g->screen_size_y);
    bto_x = MIN(g->pos_x + g->screen_size_x - 1, to_x);
    bto_y = MIN(g->pos_y + g->screen_size_y + lower_border_size - 1,to_y);
    blit(lower_border,onto,bfrom_x - g->pos_x , bfrom_y - g->pos_y - g->screen_size_y, bfrom_x,bfrom_y, bto_x - bfrom_x + 1, bto_y - bfrom_y + 1); 

    // left border
    bfrom_x = MAX(g->pos_x - left_border_size, from_x);
    bfrom_y = MAX(g->pos_y, from_y);
    bto_x = MIN(g->pos_x - 1, to_x);
    bto_y = MIN(to_y,g->pos_y + g->screen_size_y - 1);

    blit(left_border,onto,bfrom_x - g->pos_x + left_border_size, bfrom_y - g->pos_y,  bfrom_x,bfrom_y, bto_x - bfrom_x + 1, bto_y - bfrom_y + 1); 

    // right border
    bfrom_x = from_x;
    if (g->pos_x + g->screen_size_x > bfrom_x) bfrom_x = g->pos_x +g->screen_size_x;
    bfrom_y = from_y;
    if (g->pos_y > bfrom_y) bfrom_y = g->pos_y;
    bto_x = to_x;
    if (g->pos_x +g->screen_size_x + right_border_size< bto_x) bto_x = g->pos_x + g->screen_size_x + right_border_size;
    bto_y = to_y;
    if (g->pos_y + g->screen_size_y < bto_y) bto_y = g->pos_y + g->screen_size_y;
    blit(right_border,onto,bfrom_x - g->pos_x - g->screen_size_x, bfrom_y - g->pos_y,  bfrom_x,bfrom_y, bto_x - bfrom_x + 1, bto_y - bfrom_y + 1); 

    // upper left corner
    bfrom_x = MAX(from_x,g->pos_x - left_border_size);
    bfrom_y = MAX(from_y,bfrom_y - upper_border_size);
    bto_x = MIN(to_x,g->pos_x - 1);
    bto_y = MIN(to_y,g->pos_y - 1);
    blit(UpperLeftCorner,onto,bfrom_x - g->pos_x + left_border_size, bfrom_y - g->pos_y + upper_border_size, bfrom_x,bfrom_y, bto_x - bfrom_x + 1, bto_y - bfrom_y + 1); 

    // lower left corner
    bfrom_x = from_x;
    if (g->pos_x - left_border_size> bfrom_x) bfrom_x = g->pos_x - left_border_size;
    bfrom_y = from_y;
    if (g->pos_y + g->screen_size_y > bfrom_y) bfrom_y = g->pos_y +g->screen_size_y;
    bto_x = to_x;
    if (g->pos_x< bto_x) bto_x = g->pos_x;
    bto_y = to_y;
    if (g->pos_y + g->screen_size_y + lower_border_size < bto_y) bto_y = g->pos_y + g->screen_size_y + lower_border_size;
    blit(LowerLeftCorner,onto,bfrom_x - g->pos_x + left_border_size, bfrom_y - g->pos_y - g->screen_size_y, bfrom_x,bfrom_y, bto_x - bfrom_x + 1, bto_y - bfrom_y + 1); 

    // upper right corner
    bfrom_x = MAX(from_x ,g->pos_x +g->screen_size_x);
    bfrom_y = MAX(from_y,g->pos_y - upper_border_size);
    bto_x = to_x;
    if (g->pos_x +g->screen_size_x + right_border_size < bto_x) bto_x = g->pos_x +g->screen_size_x + right_border_size;
    bto_x = MIN(to_x,g->pos_x +g->screen_size_x + right_border_size - 1);
    bto_y = MIN(g->pos_y - 1, to_y);
    blit(UpperRightCorner,onto,bfrom_x - g->pos_x - g->screen_size_x, bfrom_y - g->pos_y + upper_border_size, bfrom_x,bfrom_y, bto_x - bfrom_x + 1, bto_y - bfrom_y + 1); 

    // lower right corner
    bfrom_x = from_x;
    if (g->pos_x +g->screen_size_x > bfrom_x) bfrom_x = g->pos_x +g->screen_size_x;
    bfrom_y = from_y;
    if (g->pos_y + g->screen_size_y > bfrom_y) bfrom_y = g->pos_y +g->screen_size_y;
    bto_x = to_x;
    if (g->pos_x +g->screen_size_x + right_border_size < bto_x) bto_x = g->pos_x +g->screen_size_x + right_border_size;
    bto_y = to_y;
    if (g->pos_y + g->screen_size_y + lower_border_size < bto_y) bto_y = g->pos_y + g->screen_size_y + lower_border_size;
    blit(LowerRightCorner,onto,bfrom_x - g->pos_x - g->screen_size_x, bfrom_y - g->pos_y - g->screen_size_y, bfrom_x,bfrom_y, bto_x - bfrom_x + 1, bto_y - bfrom_y + 1); 

    if (its_palette && bitmap_color_depth(onto) !=8 )
    {
        unselect_palette(its_palette);
    }

    return 0;
}






Wobject *BorderWin::process_mouse(int *x_out = NULL, int *y_out = NULL)
{
    Wobject *win;
    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y,size_was_x,size_was_y;
    int tmp_x,tmp_y;
    int flag;
    int s_pos;
    GeoInfo const *g;
    int mouse_visible = (_mouse_screen == screen);
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);
    drawing_mode(DRAW_MODE_XOR,NULL,0,0);
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    if (win != this)
    {
        set_Mouse_Cursor(WinSystem::Mouse_Cursor0);
        return win;
    }
    if (flag)
    {
        if (out_x <0 || out_y <0)
            set_Mouse_Cursor(WinSystem::Mouse_Cursor1);
        else
            set_Mouse_Cursor(WinSystem::Mouse_Cursor2);
    }
    else
        set_Mouse_Cursor(WinSystem::Mouse_Cursor0);


    g = win->get_geometric_info();
    s_pos = g->screen_size_x;
    mouse_was_x = mouse_x;
    mouse_was_y = mouse_y;
    pos_was_x = g->pos_x;
    pos_was_y = g->pos_y;
    size_was_x = g->screen_size_x;
    size_was_y = g->screen_size_y;
    tmp_x = mouse_was_x;
    tmp_y = mouse_was_y;
    if (win->get_manager()->get_parent())
        win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);

    if ((mouse_b & 2) && !flag)
    {
        pos_was_x = g->viewport_pos_x;
        pos_was_y = g->viewport_pos_y;
        while (mouse_b & 2)
        {
            mouse_is_x = mouse_x;
            mouse_is_y = mouse_y;
            if (win->get_manager()->get_parent())
                win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
    
            win->move_viewport(pos_was_x - mouse_is_x + mouse_was_x,pos_was_y - mouse_is_y + mouse_was_y);
        }
    }
    if (mouse_b & 1)
    {
        if (flag)
        {
            if (out_x < 0 || out_y <0)
            {
                                
                if (out_y <0)
                {
                    for (int i = 0;((BorderWin *)win)->systemboxes[i];i++)
                    {
                        s_pos -= ((BorderWin *)win)->systemboxes[i]->sprite->w;
                        if (out_x > s_pos)
                        {
                            if (((BorderWin *)win)->systemboxes[i]->function)
                                ((BorderWin *)win)->systemboxes[i]->function(((BorderWin *)win));
                            return NULL;
                        }
                    }
                }
              
                if (((BorderWin *)win)->Move_Style == BORDERWIN_CLEAR)
                {
                    if (mouse_visible) mouse_off();
                    rect(screen,pos_was_x, pos_was_y , pos_was_x + size_was_x  - 1, pos_was_y + size_was_y  - 1,128);
                    if (mouse_visible) mouse_on();
                }
                while (mouse_b & 1)
                {
                    mouse_is_x = mouse_x;
                    mouse_is_y = mouse_y;
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);

                    if (((BorderWin *)win)->Move_Style == BORDERWIN_CLEAR && ((tmp_x != mouse_x)||(tmp_y != mouse_y)))
                    {
                        if (mouse_visible) mouse_off();
                        rect
                        (
                            screen,
                            pos_was_x + tmp_x - mouse_was_x, 
                            pos_was_y + tmp_y - mouse_was_y, 
                            pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                            pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                            128
                        );
                        tmp_x = mouse_is_x;
                        tmp_y = mouse_is_y;
                        rect
                        (
                            screen,
                            pos_was_x + tmp_x - mouse_was_x, 
                            pos_was_y + tmp_y - mouse_was_y, 
                            pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                            pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                            128
                        );
                        if (mouse_visible) mouse_on();
                            

                    }
                    else if (((BorderWin *)win)->Move_Style == BORDERWIN_OPAQUE)
                        win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
                }

                if (((BorderWin *)win)->Move_Style == BORDERWIN_CLEAR)
                {
                    if (mouse_visible) mouse_off();
                    rect
                    (
                        screen,
                        pos_was_x + tmp_x - mouse_was_x, 
                        pos_was_y + tmp_y - mouse_was_y, 
                        pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                        pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                        128
                    );
                    if (mouse_visible) mouse_on();
                        win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
                }

            }
            else
            {
                if (((BorderWin *)win)->Move_Style == BORDERWIN_CLEAR)
                {
                    if (mouse_visible) mouse_off();
                        rect(screen,pos_was_x, pos_was_y , pos_was_x + size_was_x  - 1, pos_was_y + size_was_y  - 1,128);
                    if (mouse_visible) mouse_on();
                }

                while (mouse_b & 1)
                {
                    mouse_is_x = mouse_x;
                    mouse_is_y = mouse_y;
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);

                    if (((BorderWin *)win)->Move_Style == BORDERWIN_CLEAR && ((tmp_x != mouse_x)||(tmp_y != mouse_y)))
                    {
                        if (mouse_visible) mouse_off();
                        rect
                        (
                            screen,
                            pos_was_x,
                            pos_was_y, 
                            pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                            pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                            128
                        );
                        tmp_x = mouse_x;
                        tmp_y = mouse_y;
                        rect
                        (
                            screen,
                            pos_was_x, 
                            pos_was_y, 
                            pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                            pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                            128
                        );
                        if (mouse_visible) mouse_on();
                            

                    }
                    else if (((BorderWin *)win)->Move_Style == BORDERWIN_OPAQUE)
                        win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);
                }
                if (((BorderWin *)win)->Move_Style == BORDERWIN_CLEAR)
                {
                    if (mouse_visible) mouse_off();
                    rect
                    (
                        screen,
                        pos_was_x , 
                        pos_was_y , 
                        pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                        pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                        128
                    );
                    if (mouse_visible) mouse_on();
                        win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);
                }
            }
        }
        else
            win->raise();
    }
    solid_mode();
    if (!flag) 
        return (Wobject *)win;
    else
        return NULL;
}

RGB *BorderWin::get_its_palette()
{
    return its_palette;
}


RGB *BorderWin::its_palette = NULL;
