/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/




#include "helpers.h"
#include "boxlist.h"

int Boxlist::tagteller = 0;
int Boxlist::colorcount = 0;

Boxlist::Boxlist(Wobject *_owner,int xfrom,int yfrom, int xto, int yto, Boxlist *_last = NULL)
{
    owner = _owner;
    last = _last;
    next = NULL;
    xf = xfrom;
    yf = yfrom;
    xt = xto;
    yt = yto;
    tagteller++;
    tag= tagteller;

    if (last)
    {
        next = last->next;
        last->next = this;
        #ifdef DEBUGBOXLIST_LISTPART
        if (next)
            if (wham_trace) trace("created %d between %d en %d",tag,last->tag,next->tag);
        else
            if (wham_trace) trace("created %d after %d ",tag,last->tag);
        #endif
        if (next)
        {
            next->last = this;
        }
    }
    #ifdef DEBUGBOXLIST_LISTPART
    if (wham_trace) trace("created boxlist %d (%d,%d)-(%d,%d)",tag,xf,yf,xt,yt);
    if (last)
        trace ("after %d",last->tag);

    if (next)
        trace ("before %d",next->tag);
    #endif
}

Boxlist::~Boxlist()
{
    #ifdef DEBUGBOXLIST_LISTPART
        if (wham_trace) trace("destructor boxlist %d",tag);
    #endif
    if (next)
    {
        #ifdef DEBUGBOXLIST_LISTPART
        if (last)
            if (wham_trace) trace("reset next's last pointer to %d",last->tag);
        else
            if (wham_trace) trace("reset next's last pointer to NULL");
        #endif
        next->last = last;
    }
    if (last)
    {
        #ifdef DEBUGBOXLIST_LISTPART
          if(next)
            if (wham_trace) trace("reset last's next pointer to %d",next->tag);
          else
            if (wham_trace) trace("reset last's next pointer to NULL");

        #endif
        last->next = next;
    }
}


void Boxlist::overlay(int xfrom,int yfrom,int xto, int yto)
{



    #ifdef DEBUGBOXLIST
        int cl = screen->cl;
        int cr = screen->cr;
        int ct = screen->ct;
        int cb = screen->cb;
        mouse_off();
        set_clip(screen,0,0,SCREEN_W,SCREEN_H);
        
        BITMAP *dump = create_bitmap(SCREEN_W,SCREEN_H);
        blit(screen,dump,0,0,0,0,dump->w,dump->h);

        rectfill(screen,xfrom,yfrom,xto,yto,makecol(255,0,0));
        if (wham_trace) trace("overlay");
        rectfill(screen,xf,yf,xt,yt,makecol(0,255,0));
        if (wham_trace) trace("over this");
        blit(dump,screen,0,0,0,0,dump->w,dump->h);
        destroy_bitmap(dump);
        set_clip(screen,cl,ct,cr,cb);
        mouse_on();
    #endif


    /*
        divide the rect like:

         _______________________
        |    |  orig box        |
        |    |          3       |
        | 0  |                  |
        |    |__________________|
        |    |overlay box|      |
        |____|___________|  2   |
        |  1             |      |
        |________________|______|


    */

//    check if overlay box is inside this box, if it isn't -> skip to next
    //store original next-pointer
    Boxlist *orignext = next;

    // temporary variables , corners of the new 4 boxes and the number of new boxes
    // made ( it is possible to have 2 or 3 boxes
    int xxf[4],yyf[4],xxt[4],yyt[4], nrnew = 0;
    if (xfrom <= xt && xto >= xf && yfrom <= yt && yto >= yf)
    {
        // calculate al 4 boxes
        // box 0
        xxf[0] = xf;
        xxt[0] = MIN(xfrom - 1,xt);
        yyf[0] = yf;
        yyt[0] = MIN(yto,yt);
        // box 1
        xxf[1] = xf;
        xxt[1] = MIN(xto,xt);
        yyf[1] = MAX(yto + 1,yf);
        yyt[1] = yt;
        // box 2
        xxf[2] = MAX(xto + 1,xf);
        xxt[2] = xt;
        yyf[2] = MAX(yfrom,yf);
        yyt[2] = yt;
        // box 3
        xxf[3] = MAX(xfrom,xf);
        xxt[3] = xt;
        yyf[3] = yf;
        yyt[3] = MIN(yfrom - 1,yt);

        for (int i = 0;i<4;i++)
        {
            if (xxf[i] <= xxt[i] && yyf[i] <= yyt[i])
            {
                if (!nrnew) // reuse this box
                {
                    xf = xxf[i];
                    yf = yyf[i];
                    xt = xxt[i];
                    yt = yyt[i];
                }
                else  // make a new one
                {

                    Boxlist *n; 
                    n = new Boxlist(owner,xxf[i],yyf[i],xxt[i],yyt[i],this);
                    if (!next)
                        out_of_mem("try setting WinSystem::UseBoxes to FALSE");
                }
                nrnew++;
            }

        }
        if (!nrnew)
            xf = xt +1;
    }

    if (orignext)
        orignext->overlay(xfrom,yfrom,xto,yto);
    
    // clean_out imposiible boxes
//    for (Boxlist *i = this->next;i;i = i->next)
//    {
//        if (i->xf > i->xt || i->yf > i->yt) 
//        {
//            if (wham_trace) trace("going to delete %d %d %d %d",i->xf,i->yf,i->xt,i->yt);
//            i = i->last;
//            delete i->next;
//        }
//    }
}


void Boxlist::draw_the_boxes(BITMAP *onto,int from_x,int from_y,int to_x,int to_y)
{

    int xff,yff,xtt,ytt;
    xff = MAX(from_x,xf);
    yff = MAX(from_y,yf);
    xtt = MIN(to_x,xt);
    ytt = MIN(to_y,yt);

    #ifdef DEBUGBOXLIST
        int cl = screen->cl;
        int cr = screen->cr;
        int ct = screen->ct;
        int cb = screen->cb;
        mouse_off();
        set_clip(screen,0,0,SCREEN_W,SCREEN_H);
        
        BITMAP *dump = create_bitmap(SCREEN_W,SCREEN_H);
        blit(screen,dump,0,0,0,0,dump->w,dump->h);

        rectfill(screen,from_x,from_y,to_x,to_y,makecol(255,0,0));
        if (wham_trace) trace("area");
        rectfill(screen,xf,yf,xt,yt,makecol(0,255,0));
        if (wham_trace) trace("box");
        if ( xff <= xtt && yff <= ytt)
            rectfill(screen,xff,yff,xtt,ytt,makecol(0,0,255));
        if (wham_trace) trace("within");
        blit(dump,screen,0,0,0,0,dump->w,dump->h);
        destroy_bitmap(dump);
        set_clip(screen,cl,ct,cr,cb);
        mouse_on();
    #endif

    if ( xff <= xtt && yff <= ytt)
    {
        #ifdef SLOW
            rest(500);
        #endif

        int cl = onto->cl;
        int cr = onto->cr;
        int cb = onto->cb;
        int ct = onto->ct;
        set_clip(onto,xff,yff,xtt,ytt);

        if (owner->its_manager->_ShadowStyle)
        {
            owner->draw_shadow
            (
                onto,
                xff,
                yff,
                xtt,
                ytt
            );
        }
            

        if (owner->draw_border)
        {
            owner->draw_border
            (
                onto,
                owner,
                xff,
                yff,
                xtt,
                ytt
            );
        }       
        
        owner->draw
        (
            onto,
            xff,
            yff,
            xtt,
            ytt
        );
        set_clip(onto,cl,ct,cr,cb);

    }
    #ifdef DEBUGBOXLIST
        if (wham_trace) trace("box drawn");
    #endif
    if (next)
        next->draw_the_boxes(onto,from_x,from_y,to_x,to_y);
}


void Boxlist::draw_the_boxes(BITMAP *onto,int from_x,int from_y,int to_x,int to_y,int color)
{

    int xff,yff,xtt,ytt;
    xff = MAX(from_x,xf);
    yff = MAX(from_y,yf);
    xtt = MIN(to_x,xt);
    ytt = MIN(to_y,yt);

    if ( xff <= xtt && yff <= ytt)
    {
        #ifdef SLOW
            rest(500);
        #endif

        rectfill(onto,xff,yff,xtt,ytt,color);
    }
    if (next)
        next->draw_the_boxes(onto,from_x,from_y,to_x,to_y,color);
        

}

void Boxlist::show_the_boxes()
{
    colorcount++;
    colorcount %= 15;
    if ( xf <= xt && yf <= yt)
    {
        rectfill(screen,xf,yf,xt,yt,colorcount);
    }
    if (next)
        next->show_the_boxes();
}

void Boxlist::defrag_the_boxes()
{
    int xfs, xts, yfs , yts, found;
    // first do horizontal defrag
    for (Boxlist *j = this; j ; j = j->next)
    {
        xts = j->xf - 1;
        xfs = j->xt + 1;
        yfs = j->yf;
        yts = j->yt;

        found = TRUE;
        while (found)
        {
            found = FALSE;
            for (Boxlist *i = j->next; i ; i = i->next)
            {
                if (i->yf > i->yt || i->xf > i->xt)
                    continue;

                if (i->yf == yfs && i->yt == yts) // y coordinates equal
                {
                    if (xfs == i->xf)
                    {

                        // found one, need another pass
                        found = TRUE;
                        // reset borders of this one
                        j->xt = i->xt;
                        xfs = j->xt + 1;
                        //set back i to last so I can safely delete this one
                        i = i->last;
                        // delete it
                        delete i->next;

                    }
                    else if (xts == i->xt)
                    {
                        // found one, need another pass
                        found = TRUE;
                        // reset borders of this one
                        j->xf = i->xf;
                        xts = j->xf - 1;
                        //set back i to last so I can safely delete this one
                        i = i->last;
                        // delete it
                        delete i->next;
                    }
                } // y coords equal
            } // for boxlist i
        } //while found
    }  // for boxlist j

    // now do vertical
    for (Boxlist *j = this; j ; j = j->next)
    {
        xfs = j->xf ;
        xts = j->xt ;
        yfs = j->yt + 1;
        yts = j->yf - 1;

        found = TRUE;
        while (found)
        {
            found = FALSE;
            for (Boxlist *i = j->next; i ; i = i->next)
            {
                if (i->yf > i->yt || i->xf > i->xt)
                    continue;

                if (i->xf == xfs && i->xt == xts) // x coordinates equal
                {
                    if (yfs == i->yf)
                    {

                        // found one, need another pass
                        found = TRUE;
                        // reset borders of this one
                        j->yt = i->yt;
                        yfs = j->yt + 1;
                        //set back i to last so I can safely delete this one
                        i = i->last;
                        // delete it
                        delete i->next;

                    }
                    else if (yts == i->yt)
                    {
                        // found one, need another pass
                        found = TRUE;
                        // reset borders of this one
                        j->yf = i->yf;
                        yts = j->yf - 1;
                        //set back i to last so I can safely delete this one
                        i = i->last;
                        // delete it
                        delete i->next;
                    }
                } // y coords equal
            } // for boxlist i
        } //while found
    }  // for boxlist j

}
