/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include "guicon.h"



int GuiContainer::Border_Color = 255;
int GuiContainer::Title_Color = 0;
int GuiContainer::nr_of_instances = 0;

GuiContainer::GuiContainer(int pos_x,int pos_y, int size_x, int size_y, char *name, DIALOG *d)
      : Window(name,pos_x,pos_y,size_x,size_y,GuiContainer::draw_border)
{
    it_is_running = FALSE;
    dialog = d;
    magic_number = GUICON_MAGIC_NUMBER;
    __wham_install_dclick_check(); // installs it if not yet installed
    clear(bmp());
    set_lock_scale(TRUE);    
}

GuiContainer::~GuiContainer()
{
    __wham_remove_dclick_check();// removes it if not used anymore

}


void GuiContainer::start(int _focus_obj = -1)
{
   
   int c;
   
   res = 0;
   joy_on = TRUE;
   focus_obj = _focus_obj;

   /* initialise the dialog */
   res |= wham_dialog_message(dialog, MSG_START, 0, &obj);
   res |= wham_dialog_message(dialog, MSG_DRAW, 0, &obj, this);

   /* send show message to winsystem */
   update(); // show itself

   mouse_obj = -1; // the mouse_obj will be set correctly the first time GuiContainer::do_dialog is called



   for (c=0; dialog[c].proc; c++) {
      if (c == focus_obj)
     dialog[c].flags |= D_GOTFOCUS;
      else
     dialog[c].flags &= ~D_GOTFOCUS;
   }
   it_is_running = TRUE;
}

int GuiContainer::is_running()
{
    return it_is_running;
}



int GuiContainer::process_keyboard()
{
    do_dialog(WHAMGUI_DO_KEYS);
    return 0;
}

void GuiContainer::do_dialog(int key_or_mouse)
{
    if (it_is_running)
    {
        solid_mode();
        // all parames of next function are BY REFERENCE!! i know this is ugly, but it saves me a lot of work
        wham_do_dialog_one_step(dialog,(Window *)this,obj,res, mouse_obj,ch,joy_on, ox,oy, focus_obj,key_or_mouse);
        if (res & D_CLOSE)
            stop();
    }

}

int GuiContainer::process_message(Wobject *from, int msg)
{
    switch(msg)
    {
        case W_MSG_LOSTMOUSE:
            mouse_obj = -1;         
            return 1;
            break;
    }
    return Wobject::process_message(from,msg);
}

int GuiContainer::get_closer_object()
{
    if (!(it_is_running))
        return obj;
    else
        return -1; // is not yet closed
}

void GuiContainer::stop()
{
   /* send the finish messages */
   wham_dialog_message(dialog, MSG_END, 0, &obj);
   // send show message to window
   update();

   if (mouse_obj >= 0)
      dialog[mouse_obj].flags &= ~D_GOTMOUSE;

   hide(); // hide this window
   it_is_running = FALSE;
}


int GuiContainer::draw_border(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)
{
    int
        f_x,f_y,t_x,t_y;
    GeoInfo const *g = the_object->get_geoinfo();
    if (!onto) // return border dimensions to manager
    {
        switch(from_x)
        {
            case WINDOW_UPPER_BORDER:
                return 10;
            break;
            case WINDOW_LOWER_BORDER:
                return 3;
            break;
            case WINDOW_LEFT_BORDER:
                return 3;
            break;
            case WINDOW_RIGHT_BORDER:
                return 3;
            break;
        }

    }
    solid_mode();
    //upper border
    f_x = g->pos_x - 3;
    t_x = g->pos_x + g->screen_size_x + 2;
    f_y = g->pos_y - 10;      
    t_y = g->pos_y;           
                              
    f_x = MAX(f_x, from_x);   
    t_x = MIN(t_x, to_x);     
    f_y = MAX(f_y, from_y);   
    t_y = MIN(t_y, to_y);     
    if (f_x <= t_x && f_y <= t_y) 
        rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
        text_mode(-1);
        textout_centre(onto,font,g->name,g->pos_x + g->screen_size_x / 2, g->pos_y - 9,Title_Color); 
    //lower border                         
    f_x = g->pos_x - 3;                   
    t_x = g->pos_x + g->screen_size_x + 2; 
    f_y = g->pos_y + g->screen_size_y;     
    t_y = g->pos_y + g->screen_size_y + 2; 
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
                                           
    //left border                          
    f_x = g->pos_x - 3;                   
    t_x = g->pos_x;                        
    f_y = g->pos_y;                        
    t_y = g->pos_y + g->screen_size_y;     
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
                                           
    //right border                         
    f_x = g->pos_x + g->screen_size_x;     
    t_x = g->pos_x + g->screen_size_x + 2; 
    f_y = g->pos_y;                        
    t_y = g->pos_y + g->screen_size_y;     
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
                                           
                                           
    //END OF EXAMPLE CODE
    return 0; //returnvalue currently ignored 
}

Wobject *GuiContainer::process_mouse(int *x_out = NULL, int *y_out = NULL)
{
    Wobject *win; 
    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y,size_was_x,size_was_y; // helper variables
    int flag; // is the border clicked?
    GeoInfo const *g; // to store geoinfo from object in

    //get the object the mouse is currently in( if flag is set it is in a border)
    //out_x, out_y will be the coordinates relative to that object 
    win = WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);
    //store coordinates relative to this object if the user wants them
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    if (win != this)  return win;

    if (whamgui_raise_style == RAISE_IMMEDIATE) win->raise();
    if (whamgui_raise_style == RAISE_CLICK && mouse_b) 
    {
        win->raise();
    }

    if (!flag)
    {
        ((GuiContainer *)win)->do_dialog(WHAMGUI_DO_MOUSE);
    }
    if (flag)
    {
        if (out_x <0 || out_y <0)
            set_Mouse_Cursor(WinSystem::Mouse_Cursor1);
        else
            set_Mouse_Cursor(WinSystem::Mouse_Cursor2);
    }
    else
        set_Mouse_Cursor(WinSystem::Mouse_Cursor0);




    // mouse button 1 is currently pressed down
    if (mouse_b & 1)
    {

        g = win->get_geometric_info(); // get info on this object into g 
        mouse_was_x = mouse_x; //store current mouse position
        mouse_was_y = mouse_y;
        pos_was_x = g->pos_x; //store current object location 
        pos_was_y = g->pos_y;
        size_was_x = g->screen_size_x;//store current object size
        size_was_y = g->screen_size_y;
        
        //if the window has a parent manager, the parent window could be rescaled   
        //so calc the mouse position relative to the parent window
        if (win->get_manager()->get_parent())
             win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);
         
        if (flag) // mouse is on a border 
        {
            if (out_x < 0 || out_y <0) // left or upper border -> move
            {
                while (mouse_b & 1)
                {
                    mouse_is_x = mouse_x; // get current mouse pos
                    mouse_is_y = mouse_y;
                    // calc current mouse pos relative to parent win
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
                    // and move the window
                    win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
                }
            }
            else // left or lower border -> rescale, or move if scale_locked
            {
                while (mouse_b & 1)
                {
                    mouse_is_x = mouse_x;// get current mouse pos
                    mouse_is_y = mouse_y;
                    // make mouse pos relative to parent window
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
                    //rescale the object
                    if (!g->scale_is_locked)
                        win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);
                    else
                        win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);

                }
            }
        }
    }
    if (!flag) 
        return (Wobject *)win;
    else
        return NULL;
}

