/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <allegro.h>
#include "../include/wobject.h"
extern BITMAP * _mouse_screen;
static char buffer[1024];

#include <io.h>
#include <stdio.h>
#include <unistd.h>

FILE *DEBUG_OUTPUT_FILE = NULL;
void trace(char *format, ...);

void mouse_off()
{
    #ifdef TRACE
     if (wham_trace) trace("mouse off");
    #endif
    show_mouse(NULL);
}

void mouse_on()
{
    #ifdef TRACE
     if (wham_trace) trace("mouse on");
    #endif
    show_mouse(WHAM->TheScreen);
    if (!WHAM->TheScreen)
        show_mouse(screen);
}

void freeze_mouse()
{
    #ifdef TRACE
     if (wham_trace) trace("freeze mouse function called");
    #endif
    freeze_mouse_flag = TRUE;
}
void thaw_mouse()
{
    #ifdef TRACE
     if (wham_trace) trace("thaw mouse ");
    #endif
    freeze_mouse_flag = FALSE;
}




void out_of_mem(char *reason)
{
   WHAM = NULL;

   allegro_exit();
   static char msg[] = "Out of memory:";
   static char msg2[] = "\n\rpossibly garbage left in memory\n\r";

   _write(STDERR_FILENO, msg, sizeof(msg)-1);
   _write(STDERR_FILENO, reason, strlen(reason));
   _write(STDERR_FILENO, msg2, sizeof(msg2)-1);
    exit(1);
}

void out_of_mem()
{

   delete WHAM;

   allegro_exit();
   static char msg[] = "Out of memory:";

   _write(STDERR_FILENO, msg, sizeof(msg)-1);
    exit(1);
}


void fatal(char *format, ...)
{
    if (WHAM) delete WHAM;
    allegro_exit();
    va_list arg;
    va_start(arg,format);
    sprintf(buffer,"FATAL:");
    vsprintf(buffer+6,format,arg);
    fputs(buffer,stderr);
    va_end(arg);
    exit(1);
}

void trace(char *format, ...)
{
    int mouse_visible = (screen == _mouse_screen);
    int cl = screen->cl;
    int cr = screen->cr;
    int ct = screen->ct;
    int cb = screen->cb;
    text_mode(0);
    sprintf(buffer,"Warning: ");
    va_list arg;
    va_start(arg, format);
    vsprintf(buffer + 9,format,arg);
    va_end(arg);
    if (!DEBUG_OUTPUT_FILE)
    {        
        if (mouse_visible)
            show_mouse(NULL);
        set_clip(screen,0,0,SCREEN_W,SCREEN_H);

        text_mode(makecol(0,0,0));
        textout(screen,font,buffer,0,0,makecol(255,255,255));
        textout(screen,font,buffer,0,SCREEN_H-10,makecol(255,255,255));
        if (mouse_visible)
            show_mouse(screen);
        readkey();
        while(keypressed())
            readkey();
        if (mouse_visible)
            show_mouse(NULL);
        solid_mode();
        rectfill(screen,0,0,SCREEN_W,8,0);
        rectfill(screen,0,SCREEN_H-10,SCREEN_W,SCREEN_H,0);
        if (mouse_visible)
            show_mouse(screen);
        set_clip(screen,cl,ct,cr,cb);
    }
    else
    {
        fprintf(DEBUG_OUTPUT_FILE,buffer);
        fprintf(DEBUG_OUTPUT_FILE,"\n");
    }
        
}


void ttrace(char *format, ...)
{
    int mouse_visible = (screen == _mouse_screen);
    int cl = screen->cl;
    int cr = screen->cr;
    int ct = screen->ct;
    int cb = screen->cb;
    text_mode(0);
    sprintf(buffer,"Warning: ");
    va_list arg;
    va_start(arg, format);
    vsprintf(buffer + 9,format,arg);
    va_end(arg);
    if (!DEBUG_OUTPUT_FILE)
    {        
        if (mouse_visible)
            show_mouse(NULL);
        set_clip(screen,0,0,SCREEN_W,SCREEN_H);
        text_mode(makecol(0,0,0));
        textout(screen,font,buffer,0,0,makecol(255,255,255));
        textout(screen,font,buffer,0,SCREEN_H-10,makecol(255,255,255));
        rest(1000);
        solid_mode();
        rectfill(screen,0,0,SCREEN_W,8,0);
        rectfill(screen,0,SCREEN_H-10,SCREEN_W,SCREEN_H,0);
        if (mouse_visible)
            show_mouse(screen);
        set_clip(screen,cl,ct,cr,cb);
    }
    else
    {
        fprintf(DEBUG_OUTPUT_FILE,buffer);
        fprintf(DEBUG_OUTPUT_FILE,"\n");
    }
        
}



static Mouse_Cursor *current_Mouse_Cursor;
Mouse_Cursor *set_Mouse_Cursor(Mouse_Cursor *cursor)
{
    return NULL;
    int mouse_visible = (_mouse_screen == screen);
    Mouse_Cursor *old = current_Mouse_Cursor;
    if (current_Mouse_Cursor != cursor)
    {
        if (mouse_visible) mouse_off();
        if (!cursor)
        {
            set_mouse_sprite(NULL);
            set_mouse_sprite_focus(0,0);
        }
        else
        {
            set_mouse_sprite(cursor->bmp);
            set_mouse_sprite_focus(cursor->x,cursor->y);
        }
        if (mouse_visible) mouse_on();
        current_Mouse_Cursor = cursor;
    }
    return old;
}



void w_textout(BITMAP *b,FONT *f, char *c,int x,int y, int color,int mode)
{
    text_mode(mode);
    textout(b,f,c,x,y,color);
}

void w_textout_centre(BITMAP *b,FONT *f,char *c,int x,int y, int color,int mode)
{
    text_mode(mode);
    textout_centre(b,f,c,x,y,color);
}

static char *messages[] =
{
 "UNKOWN_MESSAGE",              //0
 "W_MSG_RAISE ",                //1
 "W_MSG_HIDE  ",                //2
 "W_MSG_SHOW  ",                //3
 "W_MSG_UPDATE",                //4
 "W_MSG_RAISED",                //5
 "W_MSG_HIDDEN",                //6
 "W_MSG_SHOWN ",                //7
 "W_MSG_MOVED ",                //8
 "W_MSG_MOVE_RESCALED",         //9
 "W_MSG_RESIZED",               //10
 "W_MSG_RESCALED",              //11
 "W_MSG_VIEWPORT_RESIZED",      //12
 "W_MSG_CREATED",               //13
 "W_MSG_DESTROYED",             //14
 "W_MSG_USER_AREA_SET",         //15
 "W_MSG_LOWERED",               //16
 "W_MSG_LOWER",                 //17
 "W_MSG_ATTACHED",              //18
 "W_MSG_SHIFTED",               //19
 "W_MSG_MANAGER_DELETE",        //20
 "W_MSG_VIEWPORT_MOVED",        //21
 "W_MSG_GOING_TO_RAISE",        //22
 "W_MSG_GOING_TO_LOWER",        //23
 "W_MSG_UPDATED",               //24
 "W_MSG_LOCKED",                //25
 "W_MSG_UNLOCKED",              //26
 "W_MSG_RAISE_AND_LOCKED",      //27
 "W_MSG_LOSTMOUSE",             //28
 "W_MSG_GOTMOUSE",              //29
 "W_MSG_FROZEN",                //30
 "W_MSG_THAWN"                  //31
};                              

char *message_string(int message)
{
    static char stdmsg[100];
    if (message > -1 || message < -31)
    {
        sprintf(stdmsg, "UNKOWN MESSAGE (%d)",message);
        return stdmsg;
    }
    return messages[-message];
}

static BITMAP *behind_progress = NULL;
static int progress_initialised = 0;

void show_progress(char *message, int now,int min, int max)
{
    static int lastx;
    int x;
    if (now>max && !progress_initialised) return;
    int xpos = SCREEN_W /2 -100;
    int ypos = SCREEN_H /2 -10;
    if (!progress_initialised)
    {
        if (!behind_progress)
            behind_progress = create_bitmap(200,20);
        if (!behind_progress)
            out_of_mem();
        blit(screen,behind_progress,xpos,ypos,0,0,200,20);
        progress_initialised = TRUE;
        rect(screen,xpos,ypos,xpos+199,ypos+19,makecol(50,50,50));
        rect(screen,xpos+1,ypos+1,xpos+199,ypos+19,makecol(150,150,150));
        rectfill(screen,xpos+1,ypos+1,xpos+198,ypos+18,makecol(100,100,100));
        lastx=0;
    }
    if (now >= max)
    {
        blit(behind_progress,screen,0,0,xpos,ypos,200,20);
        progress_initialised = FALSE;
        return;
    }
    x =  197 * (now - min)/max;
    if (x>lastx)
    {
        int mouse_visible = (screen == _mouse_screen);
        if (mouse_visible)
            mouse_off();
        rectfill (screen,xpos+2,ypos+2,xpos+2 + x,ypos+17,makecol(255,0,0));
        text_mode(-1);
        textout(screen,font,message,xpos+4,ypos+4,makecol(0,0,0));
        if (mouse_visible)
            mouse_on();
        lastx = x;
    }
}
                                             
void wham_trans_calc_progress(int pos)
{
    //textout(screen,font,"setting up shading table",0,0,15);
    //rect(screen,0,8,SCREEN_W -1, 16,15);
    //rectfill(screen,1,9,(SCREEN_W-2) * pos / 256 + 1,15,14);
    show_progress("setting up shading table",pos,0,255);
}

void remap_bitmap(BITMAP *bmp,PALETTE old, int method = 0) // currenlyt just one method, plan to add dithering later
{
    int mouse_visible = (screen == _mouse_screen);
    if (mouse_visible)
        mouse_off();
    
    int size = bmp->w * bmp->h;
    int show_progress_bar = (size > 20000);
    int col;
    int h =bmp->h;
    switch (method)
    {
    case 0:
        for (int i = 0;i<bmp->w;i++)
        {
            for (int j=0;j<bmp->h;j++)
            {
                if (show_progress_bar && !((i*h +j) % 100))
                    show_progress("remapping bitmap",i*h +j,0,size);
                col = _getpixel(bmp,i,j);
                if (col)
                {
                    _putpixel 
                    (
                        bmp,
                        i,
                        j,
                        makecol
                        (
                            old[col].r *  256/64,
                            old[col].g  * 256/64,
                            old[col].b  * 256/64
                        )
                    );
                }
            }
        }
    }
    if (show_progress_bar)
                show_progress("remapping bitmap",size,0,size);
    if (mouse_visible)
        mouse_on();
}


