/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include <stdio.h>
#include <stdarg.h>
#include "panel.h"

Panel::Panel(int size_x, int size_y,int pos_x,int pos_y,  char *name)
      : Window(name,size_x+4,size_y+4,pos_x - 2,pos_y - 2)
{
    rect(bmp(),0,0,size_x+3,size_y+3,Dark_Outline_Color);
    rect(bmp(),0,0,size_x+4,size_y+4,Light_Outline_Color);
    rectfill(bmp(),1,1,size_x+2,size_y+2,Background_Color);
    set_user_area(2,2,size_x,size_y);
    Window::background_color = Background_Color;
}

int Panel::Light_Outline_Color = 255;
int Panel::Dark_Outline_Color = 1;
int Panel::Background_Color = 0;



Wobject *Panel::process_mouse(int *x_out = NULL, int *y_out = NULL)
{
    Wobject *win; 
    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y,size_was_x,size_was_y; // helper variables
    int flag; // is the border clicked?
    GeoInfo const *g; // to store geoinfo from object in

    //get the object the mouse is currently in( if flag is set it is in a border)
    //out_x, out_y will be the coordinates relative to that object 
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);

    //store coordinates relative to this object if the user wants them
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    if (win != this)
        return win;


    // mouse button 1 is currently pressed down
    if (mouse_b & 1)
    {
       g = win->get_geometric_info(); // get info on this object into g 
       mouse_was_x = mouse_x; //store current mouse position
       mouse_was_y = mouse_y;
       pos_was_x = g->pos_x; //store current object location 
       pos_was_y = g->pos_y;
       size_was_x = g->screen_size_x;//store current object size
       size_was_y = g->screen_size_y;

       //if the window has a parent manager, the parent window could be rescaled   
       //so calc the mouse position relative to the parent window
       if (win->get_manager()->get_parent())
           win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);

       while (mouse_b & 1)
       {
            mouse_is_x = mouse_x; // get current mouse pos
            mouse_is_y = mouse_y;
            // calc current mouse pos relative to parent win
            if (win->get_manager()->get_parent())
                win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
            // and move the window
            win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
       }
       win->raise();
    }
    if (!flag) 
        return (Wobject *)win;
    else
        return NULL;
}

