/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include "helpers.h"
#include "scrolbar.h"

ScrollBar::ScrollBar(Wobject *_owner,int _type)
      : Wobject("scrollbar",10,10,0,0, _owner->get_manager())
{
    type = _type;    
    owner = _owner;
    GeoInfo const *g = owner->get_geoinfo();
    switch (type)
    {
        case TYPE_VERTICAL_SCROLLBAR:
            // both resize  and resize_s_viewport, so the object really gets this size
            // both on screen and internal

            resize(10,g->screen_size_y + g->Upper_Border_Size + g->Lower_Border_Size);
            resize_s_viewport(10,g->screen_size_y + g->Upper_Border_Size + g->Lower_Border_Size);
            attach_to(owner,WINDOW_UL_CORNER,WINDOW_UR_CORNER,0,0);
        break;
        case TYPE_HORIZONTAL_SCROLLBAR:
            // both resize  and resize_s_viewport, so the object really gets this size
            // both on screen and internal
            resize(g->screen_size_x + g->Left_Border_Size + g->Right_Border_Size, 10);
            resize_s_viewport(g->screen_size_x + g->Left_Border_Size + g->Right_Border_Size, 10);
            attach_to(owner,WINDOW_UL_CORNER,WINDOW_LL_CORNER,0,0);
        break;
    }
    g = get_geoinfo();
}

int ScrollBar::process_message(Wobject  *from,int message)
{
    
    GeoInfo const *g = owner->get_geoinfo();
    if (from == owner)
    {
        if (message == W_MSG_SHOWN)
        {
            if (g->size_x != g->viewport_size_x || g->size_y != g->viewport_size_y)   
                show();
            return TRUE;
        }
        else if (message == W_MSG_HIDDEN)
        {
            hide();
            return TRUE;
        }
    }
    int ret = Wobject::process_message(from,message);
    if (from == owner)
    {
        switch(message)
        {
            case W_MSG_VIEWPORT_RESIZED:
                //fall through
            case W_MSG_MOVE_RESCALED:
                //fall through
            case W_MSG_RESCALED:
                //fall through
            case W_MSG_RESIZED:
                //fall through
             switch (type)
             {
                 case TYPE_VERTICAL_SCROLLBAR:
                     resize(10,g->screen_size_y + g->Upper_Border_Size + g->Lower_Border_Size);
                     resize_s_viewport(10,g->screen_size_y + g->Upper_Border_Size + g->Lower_Border_Size);
                 break;
                 case TYPE_HORIZONTAL_SCROLLBAR:
                     resize(g->screen_size_x + g->Left_Border_Size + g->Right_Border_Size, 10);
                     resize_s_viewport(g->screen_size_x + g->Left_Border_Size + g->Right_Border_Size, 10);
                 break;
             }
             switch(type)
             {

                 case TYPE_VERTICAL_SCROLLBAR:
                     if (g->size_y == g->viewport_size_y)
                     {
                        hide();
                     }
                     else if (g->is_displayed)
                     {
                        show();
                     }
                 break;
                 case TYPE_HORIZONTAL_SCROLLBAR:
                     if (g->size_x == g->viewport_size_x)
                     {
                        hide();
                     }
                     else if (g->is_displayed)
                     {
                        show();
                     }
                 
                 break;
             }
            break;
            case W_MSG_VIEWPORT_MOVED:
             update();
            break;
        }       

    }
    return ret;                

}


void ScrollBar::draw(BITMAP *onto, int f_x, int f_y,int t_x, int t_y)
{
      GeoInfo const *gowner = owner->get_geoinfo(),
                    *g = get_geoinfo();
      solid_mode();
      // for now relies on overlying clipping so silence compiler about unused vars
        f_x = f_y = t_x = t_y = 0;
      //
        rect(onto,g->pos_x,g->pos_y,g->pos_x + g->screen_size_x - 1,g->pos_y + g->screen_size_y - 1, Light_Color);
        rect(onto,g->pos_x + 1,g->pos_y + 1,g->pos_x + g->screen_size_x - 1,g->pos_y + g->screen_size_y - 1, Dark_Color);
        rect(onto,g->pos_x + 2,g->pos_y + 2,g->pos_x + g->screen_size_x - 2,g->pos_y + g->screen_size_y - 2, Light_Color);
        rectfill(onto,g->pos_x + 2,g->pos_y + 2,g->pos_x + g->screen_size_x - 3,g->pos_y + g->screen_size_y - 3, Middle_Color);

        switch(type)
        {
        case TYPE_VERTICAL_SCROLLBAR:
            bar_length = g->screen_size_y - 4;
            marker_length = (bar_length * gowner->viewport_size_y) / gowner->size_y;
            marker_pos = ((bar_length+1) * gowner->viewport_pos_y) / gowner->size_y;
            rect(onto,g->pos_x + 2, g->pos_y + 2 + marker_pos, g->pos_x + g->screen_size_x -3, g->pos_y + 1 + marker_length + marker_pos, Light_Color);
            rect(onto,g->pos_x + 3, g->pos_y + 3 + marker_pos, g->pos_x + g->screen_size_x -3, g->pos_y + 1 + marker_length + marker_pos, Dark_Color);
            rectfill(onto,g->pos_x + 3, g->pos_y + 3 + marker_pos, g->pos_x + g->screen_size_x -4, g->pos_y + marker_length + marker_pos, Middle_Color);
            break;
        case TYPE_HORIZONTAL_SCROLLBAR:
            bar_length = g->screen_size_x - 4;
            marker_length = (bar_length * gowner->viewport_size_x) / gowner->size_x;
            marker_pos = ((bar_length+1) * gowner->viewport_pos_x) / gowner->size_x;
            rect(onto,g->pos_x + 2 + marker_pos, g->pos_y + 2, g->pos_x + marker_pos + 1 + marker_length, g->pos_y + g->screen_size_y - 3, Light_Color);
            rect(onto,g->pos_x + 3 + marker_pos, g->pos_y + 3, g->pos_x + marker_pos + 1 + marker_length, g->pos_y + g->screen_size_y - 3, Dark_Color);
            rectfill(onto,g->pos_x + 3 + marker_pos, g->pos_y + 3 , g->pos_x + marker_pos + marker_length, g->pos_y + g->screen_size_y - 4, Middle_Color);
            break;
        }
}

// MOUSE FUNCTION EDIT FOR DESIRED ACTIONS
Wobject *ScrollBar::process_mouse(int *x_out = NULL, int *y_out = NULL)
{
    Wobject *win; 
    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y; // helper variables
    int flag; // is the border clicked?
    GeoInfo const *g; // to store geoinfo from object in

    //get the object the mouse is currently in( if flag is set it is in a border)
    //out_x, out_y will be the coordinates relative to that object 
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);

    //store coordinates relative to this object if the user wants them
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    if (win != this)
        return win;

    set_Mouse_Cursor(WinSystem::Mouse_Cursor0);
    double scale;
    // mouse button 1 is currently pressed down
    if (mouse_b & 1)
    {
       g = win->get_geometric_info(); // get info on this object into g 

       switch(type)
       {
        case TYPE_VERTICAL_SCROLLBAR:
         if (out_x > 2 && out_x < g->screen_size_x - 2)
         {
            scale = (double)(owner->get_geoinfo()->size_y)/(double)bar_length;
            // if on marker
            pos_was_x = owner->get_geoinfo()->viewport_pos_x;
            pos_was_y = owner->get_geoinfo()->viewport_pos_y;

            if (out_y > marker_pos + 2 && out_y < marker_pos + marker_length + 2)
            {
                mouse_was_x = mouse_x;
                mouse_was_y = mouse_y;
                if (win->get_manager()->get_parent())
                    win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);
                while(mouse_b & 1)
                {
                    mouse_is_x = mouse_x;
                    mouse_is_y = mouse_y;
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
                    owner->move_viewport(pos_was_x,(int)(pos_was_y + (mouse_is_y - mouse_was_y) * scale));
                }
            }
            else if (out_y < marker_pos + 2 )
            {
               while (mouse_b & 1);
               owner->move_viewport(pos_was_x,(int)(pos_was_y - (owner->get_geoinfo()->screen_size_y/1.5) / owner->get_geoinfo()->scale_y)) ;
            }
            else if (out_y > marker_pos + marker_length + 2 )
            {
               while (mouse_b & 1);
               owner->move_viewport(pos_was_x,(int)(pos_was_y + (owner->get_geoinfo()->screen_size_y/1.5) / owner->get_geoinfo()->scale_y));
            }
         }
         break;
        case TYPE_HORIZONTAL_SCROLLBAR:
         if (out_y > 2 && out_y < g->screen_size_y - 2)
         {
            scale = (double)(owner->get_geoinfo()->size_x)/(double)bar_length;
            pos_was_x = owner->get_geoinfo()->viewport_pos_x;
            pos_was_y = owner->get_geoinfo()->viewport_pos_y;

            // if on marker
            if (out_x > marker_pos + 2 && out_x < marker_pos + marker_length)
            {
                mouse_was_x = mouse_x;
                mouse_was_y = mouse_y;
                if (win->get_manager()->get_parent())
                    win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);
                while(mouse_b & 1)
                {
                    mouse_is_x = mouse_x;
                    mouse_is_y = mouse_y;
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
                    owner->move_viewport((int)(pos_was_x + (mouse_is_x - mouse_was_x) * scale),pos_was_y);
                }
            }
            else if (out_x < marker_pos + 2 )
            {
               while (mouse_b & 1);
               owner->move_viewport((int)(pos_was_x - (owner->get_geoinfo()->screen_size_x/1.5) / owner->get_geoinfo()->scale_x),pos_was_y) ;
            }
            else if (out_x > marker_pos + marker_length )
            {
               while (mouse_b & 1);
               owner->move_viewport((int)(pos_was_x + (owner->get_geoinfo()->screen_size_x/1.5) /owner->get_geoinfo()->scale_x),pos_was_y) ;
            }

         }
         break;
       }

    }
       return (Wobject *)win;
}

int ScrollBar::White_Color = 251;
int ScrollBar::Light_Color = 252;
int ScrollBar::Middle_Color = 253;
int ScrollBar::Dark_Color = 254;
int ScrollBar::Black_Color = 255;

