/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#include "helpers.h"
#include "secretar.h"
#include <screensv.h>

int Secretary::initialized = 0;

void Secretary::destroy()
{
    if (initialized)
    {
        if (screensaver)
            set_screen_saver(NULL,0);
        if (schedulerlist)
        {
            while(schedulerlist->next)
                delete schedulerlist->next;
            delete schedulerlist;
        }
        initialized = 0;
    }
    if (number_of_loaded_screensavers)
    {
        for (int i=0;i<number_of_loaded_screensavers;i++)
        {
            delete loaded_screensaver_filenames[i];
        }
        free(loaded_screensaver_filenames);
        free(loaded_screensaver_names);
        free(loaded_screensavers);
        number_of_loaded_screensavers = 0;
        loaded_screensaver_names = NULL;
        loaded_screensavers = NULL;
    }
}

void Secretary::create()
{
    initialized++;
    if (!schedulerlist)
    {
        schedulerlist = new Scheduler_List(NULL,NULL,0,NULL);
    }    
}

Secretary::Secretary()
{
    create();

}

Secretary::~Secretary()
{
    destroy();
}

int Secretary::number_of_loaded_screensavers = 0;
char **Secretary::loaded_screensaver_names = NULL;
char **Secretary::loaded_screensaver_filenames = NULL;
void **Secretary::loaded_screensavers = NULL;
char *Secretary::current_screen_saver_name = NULL;

char *Secretary::get_current_screen_saver_name()
{
    return current_screen_saver_name;
}

Wobject *Secretary::mouse_object = NULL;

Wobject *Secretary::process_mouse(int *x = NULL, int *y = NULL)
{
    int out_x,out_y,flag; // not used but needed for process_coord
    Wobject *obj = WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);
           
    if (screensaver)
    {
        if (keypressed() || (old_mouse_x !=mouse_x) || (old_mouse_y != mouse_y) || mouse_b || screensaver_suspended)
        {
               ScreenScounter = 0;
        }
        old_mouse_x = mouse_x;
        old_mouse_y = mouse_y;

        if (ScreenScounter>ScreenStime)
        {
            PALETTE p;
            get_palette(p);
            mouse_off();
            screensaver();
            set_palette(p);
            WHAM->refresh_all_visible_objects();
            mouse_on();
            ScreenScounter = 0;
        }
    }
    schedulerlist->do_list();

    if (obj != mouse_object)
    {
        if (mouse_object)
        {
            WHAM->send_message(NULL,W_MSG_LOSTMOUSE,mouse_object);
        }
        mouse_object = obj;
        if (mouse_object)
        {
            WHAM->send_message(NULL,W_MSG_GOTMOUSE,mouse_object);
        }
    }

    if (WHAM->get_keyboard_claimer() && keypressed())
    {
	    WHAM->get_keyboard_claimer()->process_keyboard();
       ScreenScounter = 0;
    }


    if (obj)
    {
        if (!(WHAM->get_keyboard_claimer()) && keypressed())
        {
            obj->process_keyboard();
            ScreenScounter = 0;
        }
        else          // else is neccesary becaus the keys could have closed the window
            return obj->process_mouse(x, y);
    }
    
    if (catch_keys)
        while (keypressed())        //read away unused keys
        {
            readkey();
            ScreenScounter= 0;
        }
    
    return NULL;
}

void ScreenStimer(...) //just increases some counters every second
{
        Secretary::SecCounter++;
        Secretary::SecCounter %= 360000;
        Secretary::ScreenScounter++;
        Secretary::ScreenScounter %= 32000;
}
END_OF_FUNCTION(ScreenStimer);


void Secretary::set_screen_saver(void (*_screensaver)(),int time, char *name = NULL)
{

        if (!init_timer)
        {
            LOCK_VARIABLE(ScreenScounter);
            LOCK_VARIABLE(SecCounter);
            LOCK_FUNCTION(ScreenStimer);
            install_int(ScreenStimer,1000);
            init_timer++;
        }
        screensaver = _screensaver;
        ScreenStime = time;
        if (!name)
            current_screen_saver_name = "User defined";
        else
            current_screen_saver_name = name;


}

int Secretary::load_screen_saver(char *name,int time)
{
        screensaver = NULL;
        char *screensaver_name = NULL;
        for (int i=0;i<number_of_loaded_screensavers;i++)
        {
            if (!strcmp(name,loaded_screensaver_names[i]))
                screensaver = (void (*)())loaded_screensavers[i];
        }
        // since screensavers cannot be unloaded, I keep a list of
        // already loaded screensavers, to prevent loading them twice
        // which would be a waste of memory
        if (!screensaver)
        {
            screensaver = load_screensaver(name,&screensaver_name);
            if (screensaver)
            {
                number_of_loaded_screensavers++;
                loaded_screensaver_names =(char **)realloc(loaded_screensaver_names, sizeof(char *) * number_of_loaded_screensavers);
                loaded_screensaver_filenames =(char **)realloc(loaded_screensaver_names, sizeof(char *) * number_of_loaded_screensavers);
                loaded_screensavers = (void **)realloc(loaded_screensavers, sizeof(void *) * number_of_loaded_screensavers);
                loaded_screensaver_filenames[number_of_loaded_screensavers - 1] = strdup(name);
                current_screen_saver_name = loaded_screensaver_names[number_of_loaded_screensavers - 1] = screensaver_name;
                loaded_screensavers[number_of_loaded_screensavers - 1] = (void *)screensaver;
                
            }
            else
                return -1;

        }


        if (!init_timer)
        {
            LOCK_VARIABLE(ScreenScounter);
            LOCK_VARIABLE(SecCounter);
            LOCK_FUNCTION(ScreenStimer);
            install_int(ScreenStimer,1000);
            init_timer++;
        }
        ScreenStime = time;
        return 0;
}

void Secretary::reset_screen_saver_time(int time)
{
        ScreenStime = time;
}

void Secretary::set_key_catching(int _catch)
{
    catch_keys = _catch;
}

void Secretary::set_screensaver_suspension(int susp)
{
    screensaver_suspended = susp;
}

void Secretary::schedule(int (*fun)(void *arg),void *arg,int itime)
{
    if (!init_timer)
    {
        LOCK_VARIABLE(ScreenScounter);
        LOCK_VARIABLE(SecCounter);
        LOCK_FUNCTION(ScreenStimer);
        install_int(ScreenStimer,1000);
        init_timer++;
    }
    // add one to the list by creating it
    Scheduler_List *n; 
        n = new Scheduler_List(fun,arg,itime,schedulerlist);

}


int Secretary::ScreenScounter = 0;
int Secretary::SecCounter = 0;
int Secretary::ScreenStime = 0;
int Secretary::init_timer = 0;

int Secretary::screensaver_suspended= FALSE;

int Secretary::catch_keys = TRUE;

void (*Secretary::screensaver)() = NULL;

int Secretary::old_mouse_x = 0;
int Secretary::old_mouse_y = 0;

Scheduler_List *Secretary::schedulerlist = NULL;


Scheduler_List::Scheduler_List(int (*f)(void *arg),void *_arg,int t,Scheduler_List *head)
{
    if (head)
    {
        next = head->next;
        last = head;
        head->next = this;
        if (next)
            next->last = this;
    }
    else
    {
        next = NULL;
        last = NULL;
    }
    function = f;
    time_increment = t;
    next_execute_time = Wham_Secretary.SecCounter;
    arg = _arg;
}


Scheduler_List::~Scheduler_List()
{
    if (next)
        next->last = last;
    if (last)
        last->next =next;
}
// execute the function of the next, so the first elem never gets executed
// this is so because then I'm able to delete it
void Scheduler_List::do_list()
{
    if (next && next->function && next->next_execute_time <= Wham_Secretary.SecCounter)
    {
        if (next->function(next->arg))
        {
            delete next;    // if returns nonzero, finish

        }
        else
            next->next_execute_time += next->time_increment;
    }
}



//**********************************

Secretary Wham_Secretary;




