/*
 *         gui routines altered for WHAM
 *         this file is nearly literally ripped from Allegro
 *         the main changes needed were because I needed the gui routines
 *         to draw on another bitmap than screen
 *         and I needed to rescale the mouse coordinates
 *         and some helper functions are added
 *         this file still needs the original GUI.c linked in for some globals
 *         I'll fix that ASAP.
 *         Martijn Versteegh
 *         p.s. I'd like to thank Shawn and C La Mantia for having such a 
 *         clear programming style ( it took me about 1 day to convert this 
 *         file which is not much if you consider the murky structure of WHAM)
 */
/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *      By Shawn Hargreaves,
 *      1 Salisbury Road,
 *      Market Drayton,
 *      Shropshire,
 *      England, TF9 1AJ.
 *
 *      GUI routines.
 *
 *      Radio button object by Chris La Mantia.
 *
 *      See readme.txt for copyright information.(in Allegro directory, that is)
 */


#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <dir.h>
#include <limits.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include "window.h"
#include "whamgui.h"

#ifndef _USE_LFN
#define _USE_LFN  0
#endif


int whamgui_raise_style = RAISE_IMMEDIATE;

/*the screen it is drawn on */
BITMAP * wham_guiscreen = NULL;

/*the object it is drawn on */
static Window * wham_guiobject = NULL;

int WHAM_SEND_MESSAGE(DIALOG *d,int msg,int c)  
{
	int ret = (d)->proc(msg, d, c);
	if (msg == MSG_DRAW)
	{
		wham_guiobject->update();
	}
	return ret;
}

int whamgui_light_color = 15;
int whamgui_middle_color =4;
int whamgui_dark_color = 0;
int whamgui_black_color = 0;
int whamgui_white_color = 255;
int whamgui_Text_Color = 0;

/* pointer to the currently active dialog */
static DIALOG *the_dialog = NULL;


typedef char *(*getfuncptr)(int, int *);

/*
 * sets the screen on which the dialog is drawn
 */
static Window * __wham_set_gui_container(Window *the_object)
{
	Window *old = wham_guiobject; 
	if (the_object)
		wham_guiscreen = the_object->bmp();
	else
		wham_guiscreen = NULL;

	wham_guiobject = the_object;
	return old;
}

/*
 * substituions for mous_x and mouse_y
 */
inline int __wham_mouse_x()
{
	int 
		xout = mouse_x,
		yout = mouse_y;
	wham_guiobject->transform_coordinates(&xout,&yout);
	return xout;
}

inline int __wham_mouse_y()
{
	int 
		xout = mouse_x,
		yout = mouse_y;
	wham_guiobject->transform_coordinates(&xout,&yout);
	return yout;
}

/* gui_textout:
 *  Wrapper function for drawing text to the screen, which interprets the
 *  & character as an underbar for displaying keyboard shortcuts. Returns
 *  the width of the output string in pixels.
 */
static int gui_textout(BITMAP *bmp, char *s, int x, int y, int color, int centre)
{
   char tmp[160];
   int c;
   int len = 0;
   int pix_len;
   int hline_pos = -1;

   for (c=0; (s[c]) && (len<159); c++) {
	  if (s[c] == '&') {
	 if (s[c+1] != '&')
		hline_pos = len;
	 else {
		tmp[len++] = '&';
		c++;
	 }
	  }
	  else
	 tmp[len++] = s[c];
   }

   tmp[len] = 0;
   pix_len = text_length(font, tmp);

   if (centre)
	  x -= pix_len / 2;

   if (bmp) {
	  textout(bmp, font, tmp, x, y, color);

	  if (hline_pos >= 0) {
	 c = tmp[hline_pos];
	 tmp[hline_pos] = 0;
	 hline_pos = text_length(font, tmp);
	 tmp[0] = c;
	 tmp[1] = 0;
	 c = text_length(font, tmp);
	 hline(bmp, x+hline_pos, y+text_height(font)-gui_font_baseline, x+hline_pos+c-1, color);
	  }
   }

   return pix_len;
}



/* gui_strlen:
 *  Returns the length of a string in pixels, ignoring '&' characters.
 */
static int gui_strlen(char *s)
{
   return gui_textout(NULL, s, 0, 0, 0, 0);
}



/* dotted_rect:
 *  Draws a dotted rectangle, for showing an object has the input focus.
 */
static void dotted_rect(int x1, int y1, int x2, int y2, int fg, int bg)
{
   int c;

   for (c=x1; c<x2; c+=2) {
	  putpixel(wham_guiscreen, c, y1, fg);
	  putpixel(wham_guiscreen, c, y2, fg);
   }

   for (c=x1+1; c<x2; c+=2) {
	  putpixel(wham_guiscreen, c, y1, bg);
	  putpixel(wham_guiscreen, c, y2, bg);
   }

   for (c=y1; c<y2; c+=2) {
	  putpixel(wham_guiscreen, x1, c, fg);
	  putpixel(wham_guiscreen, x2, c, fg);
   }

   for (c=y1+1; c<y2; c+=2) {
	  putpixel(wham_guiscreen, x1, c, bg);
	  putpixel(wham_guiscreen, x2, c, bg);
   }
}



/* d_clear_proc:
 *  Simple dialog procedure which just clears the screen. Useful as the
 *  first object in a dialog.
 */
int wham_d_clear_proc(int msg, DIALOG *d, int c)
{
   c = 0; /* to stop the compiler from complaining */
   if (msg == MSG_DRAW) {
	  clear_to_color(wham_guiscreen, d->bg);
   }

   return D_O_K;
}



/* d_box_proc:
 *  Simple dialog procedure: just draws a box.
 */
int wham_d_box_proc(int msg, DIALOG *d, int c)
{
   c = 0; /* to stop the compiler from complaining */
   if (msg==MSG_DRAW) {
	  int fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	  rectfill(wham_guiscreen, d->x+1, d->y+1, d->x+d->w-1, d->y+d->h-1, d->bg);
	  rect(wham_guiscreen, d->x, d->y, d->x+d->w, d->y+d->h, fg);
   }

   return D_O_K;
}


int wham_d_3dbox_proc(int msg, DIALOG *d, int c)
{
   c = 0; /* to stop the compiler from complaining */
   if (msg==MSG_DRAW) {
	  if (d->fg)
	  {
		rectfill(wham_guiscreen, d->x+1, d->y+1, d->x+d->w-1, d->y+d->h-1, whamgui_middle_color);
		vline(wham_guiscreen,d->x+1,d->y+1,d->y+d->h-1,whamgui_light_color);
		vline(wham_guiscreen,d->x+2,d->y+2,d->y+d->h-2,whamgui_light_color);
		hline(wham_guiscreen,d->x+1,d->y+1,d->x+d->w-1,whamgui_light_color);
		hline(wham_guiscreen,d->x+2,d->y+2,d->x+d->w-2,whamgui_light_color);
		hline(wham_guiscreen,d->x+1,d->y+d->h-1,d->x+d->w-1,whamgui_dark_color);
		hline(wham_guiscreen,d->x+2,d->y+d->h-2,d->x+d->w-2,whamgui_dark_color);
		vline(wham_guiscreen,d->x+d->w-1,d->y+1,d->y+d->h-1,whamgui_dark_color);
		vline(wham_guiscreen,d->x+d->w-2,d->y+2,d->y+d->h-2,whamgui_dark_color);
	  }
	  else
	  {
		rectfill(wham_guiscreen, d->x+1, d->y+1, d->x+d->w-1, d->y+d->h-1, whamgui_middle_color);
		vline(wham_guiscreen,d->x+1,d->y+1,d->y+d->h-1,whamgui_dark_color);
		vline(wham_guiscreen,d->x+2,d->y+2,d->y+d->h-2,whamgui_dark_color);
		hline(wham_guiscreen,d->x+1,d->y+1,d->x+d->w-1,whamgui_dark_color);
		hline(wham_guiscreen,d->x+2,d->y+2,d->x+d->w-2,whamgui_dark_color);
		hline(wham_guiscreen,d->x+1,d->y+d->h-1,d->x+d->w-1,whamgui_light_color);
		hline(wham_guiscreen,d->x+2,d->y+d->h-2,d->x+d->w-2,whamgui_light_color);
		vline(wham_guiscreen,d->x+d->w-1,d->y+1,d->y+d->h-1,whamgui_light_color);
		vline(wham_guiscreen,d->x+d->w-2,d->y+2,d->y+d->h-2,whamgui_light_color);
	  }

	  rect(wham_guiscreen, d->x, d->y, d->x+d->w, d->y+d->h, whamgui_black_color);
   }

   return D_O_K;
}





/* d_shadow_box_proc:
 *  Simple dialog procedure: draws a box with a shadow.
 */
int wham_d_shadow_box_proc(int msg, DIALOG *d, int c)
{
   c = 0; /* to stop the compiler from complaining */
   if (msg==MSG_DRAW) {
	  int fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	  rectfill(wham_guiscreen, d->x+1, d->y+1, d->x+d->w-2, d->y+d->h-2, d->bg);
	  rect(wham_guiscreen, d->x, d->y, d->x+d->w-1, d->y+d->h-1, fg);
	  vline(wham_guiscreen, d->x+d->w, d->y+1, d->y+d->h, fg);
	  hline(wham_guiscreen, d->x+1, d->y+d->h, d->x+d->w, fg);
   }

   return D_O_K;
}



/* d_bitmap_proc:
 *  Simple dialog procedure: draws the bitmap which is pointed to by dp.
 */
int wham_d_bitmap_proc(int msg, DIALOG *d, int c)
{
   BITMAP *b = (BITMAP *)d->dp;

   if (msg==MSG_DRAW)
	  blit(b, wham_guiscreen, 0, 0, d->x, d->y, d->w, d->h);

   c = 0; /* to stop the compiler from complaining */
   return D_O_K;
}



/* d_text_proc:
 *  Simple dialog procedure: draws the text string which is pointed to by dp.
 */
int wham_d_text_proc(int msg, DIALOG *d, int c)
{
   if (msg==MSG_DRAW) {
	  int fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	  text_mode(d->bg);
	  gui_textout(wham_guiscreen,(char *)d->dp, d->x, d->y, fg, FALSE);
   }

   c = 0; /* to stop the compiler from complaining */
   return D_O_K;
}



/* d_ctext_proc:
 *  Simple dialog procedure: draws the text string which is pointed to by dp,
 *  centering it around the object's x coordinate.
 */
int wham_d_ctext_proc(int msg, DIALOG *d, int c)
{
   if (msg==MSG_DRAW) {
	  int fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	  text_mode(d->bg);
	  gui_textout(wham_guiscreen,(char *)d->dp, d->x, d->y, fg, TRUE);
   }

   c = 0; /* to stop the compiler from complaining */
   return D_O_K;
}



/* d_button_proc:
 *  A button object (the dp field points to the text string). This object
 *  can be selected by clicking on it with the mouse or by pressing its 
 *  keyboard shortcut. If the D_EXIT flag is set, selecting it will close 
 *  the dialog, otherwise it will toggle on and off.
 */
int wham_d_button_proc(int msg, DIALOG *d, int c)
{
   int state1, state2;
   int swap;
   int g;
   c = 0; /* stop the compiler from complaining 'bout unused vars */

   switch (msg) {

	  case MSG_DRAW:
	 if (d->flags & D_SELECTED) {
		g = 1;
		state1 = d->bg;
		state2 = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	 }
	 else {
		g = 0; 
		state1 = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
		state2 = d->bg;
	 }

	 rectfill(wham_guiscreen, d->x+1+g, d->y+1+g, d->x+d->w-2+g, d->y+d->h-2+g, state2);
	 rect(wham_guiscreen, d->x+g, d->y+g, d->x+d->w-1+g, d->y+d->h-1+g, state1);
	 text_mode(state2);
	 gui_textout(wham_guiscreen, (char *)d->dp, d->x+d->w/2+g, d->y+d->h/2-text_height(font)/2+g, state1, TRUE);

	 if (d->flags & D_SELECTED) {
		vline(wham_guiscreen, d->x, d->y, d->y+d->h-1, d->bg);
		hline(wham_guiscreen, d->x, d->y, d->x+d->w-1, d->bg);
	 }
	 else {
		vline(wham_guiscreen, d->x+d->w, d->y+1, d->y+d->h-1, d->fg);
		hline(wham_guiscreen, d->x+1, d->y+d->h, d->x+d->w, d->fg);
	 }
	 if ((d->flags & D_GOTFOCUS) && 
		 (!(d->flags & D_SELECTED) || !(d->flags & D_EXIT)))
		dotted_rect(d->x+1+g, d->y+1+g, d->x+d->w-2+g, d->y+d->h-2+g, state1, state2);
	 break;

	  case MSG_WANTFOCUS:
	 return D_WANTFOCUS;

	  case MSG_KEY:
	 /* close dialog? */
	 if (d->flags & D_EXIT)
		return D_CLOSE;

	 /* or just toggle */
	 d->flags ^= D_SELECTED;
	 WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
	 break;

	  case MSG_CLICK:

	 /* what state was the button originally in? */
	 state1 = d->flags & D_SELECTED;
	 if (d->flags & D_EXIT)
		swap = FALSE;
	 else
		swap = state1;

	 /* track the mouse until it is released */
	 while (mouse_b) {
		state2 = ((__wham_mouse_x() >= d->x) && (__wham_mouse_y() >= d->y) &&
			 (__wham_mouse_x() <= d->x + d->w) && (__wham_mouse_y() <= d->y + d->h));
		if (swap)
		   state2 = !state2;

		/* redraw? */
		if (((state1) && (!state2)) || ((state2) && (!state1))) {
		   d->flags ^= D_SELECTED;
		   state1 = d->flags & D_SELECTED;
		   WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
		}
	 }

	 /* should we close the dialog? */
	 if ((d->flags & D_SELECTED) && (d->flags & D_EXIT)) {
		d->flags ^= D_SELECTED;
		return D_CLOSE;
	 }
	 break; 
   }

   return D_O_K;
}

/* d_3dbutton_proc:
 *  A button object (the dp field points to the text string). This object
 *  can be selected by clicking on it with the mouse or by pressing its 
 *  keyboard shortcut. If the D_EXIT flag is set, selecting it will close 
 *  the dialog, otherwise it will toggle on and off.
 *  it ignores the fg and bg variables and uses the global 
 *  whamgui_*_color variables and gui_mg_color;
 *  
 */
int wham_d_3dbutton_proc(int msg, DIALOG *d, int c)
{

	if (msg == MSG_DRAW)
	{
		int     
			color1 = whamgui_light_color,
			color2 = whamgui_middle_color,
			color3 = whamgui_dark_color,
			colortext = (d->flags & D_DISABLED) ? gui_mg_color : whamgui_Text_Color;

		if (d->flags & D_SELECTED) 
		{   
			color1 = whamgui_dark_color;
			color2 = whamgui_middle_color;
			color3 = whamgui_light_color;
		}
	   
		rectfill(wham_guiscreen, d->x+2, d->y+2, d->x+d->w-3, d->y+d->h-3, color2);
		vline(wham_guiscreen, d->x + 1, d->y + 1, d->y+d->h-2,color1);
		hline(wham_guiscreen, d->x + 1, d->y + 1, d->x+d->w-2,color1);
		vline(wham_guiscreen, d->x + d->w - 2, d->y +  1, d->y + d->h - 2, color3);
		hline(wham_guiscreen, d->x + 1, d->y + d->h - 2, d->x + d->w - 2, color3);
		vline(wham_guiscreen, d->x + 2, d->y + 2, d->y+d->h-3,color1);
		hline(wham_guiscreen, d->x + 2, d->y + 2, d->x+d->w-3,color1);
		vline(wham_guiscreen, d->x + d->w - 3, d->y +  2, d->y + d->h - 3, color3);
		hline(wham_guiscreen, d->x + 2, d->y + d->h - 3, d->x + d->w - 3, color3);
		rect(wham_guiscreen,d->x,d->y, d->x+d->w-1, d->y+d->h-1 ,whamgui_black_color);
//        vline(wham_guiscreen, d->x, d->y, d->y+d->h-1,color1);
//        hline(wham_guiscreen, d->x, d->y, d->x+d->w-1,color1);
//        vline(wham_guiscreen, d->x +d->w - 1, d->y, d->y+d->h-1, color3);
//        hline(wham_guiscreen, d->x, d->y+d->h - 1, d->x+d->w-1, color3);

		text_mode(color2);
		gui_textout
		(
			wham_guiscreen, 
			(char *)d->dp, 
			d->x+d->w/2, 
			d->y+d->h/2-text_height(font)/2, 
			colortext, 
			TRUE
		);

		if ((d->flags & D_GOTFOCUS) && 
			(!(d->flags & D_SELECTED) || !(d->flags & D_EXIT)))
		{
			dotted_rect
			(
				d->x+d->w/2 - gui_strlen((char *)d->dp) / 2 - 1, 
				d->y+d->h/2 - text_height(font)/2 - 1, 
				d->x+d->w/2 + gui_strlen((char *)d->dp) / 2 + 1, 
				d->y+d->h/2 + text_height(font)/2+1 , 
				color2,
				colortext
			);
		}

		return D_O_K;
	}
	return wham_d_button_proc(msg,d,c);
}


/* d_check_proc:
 *  Who needs C++ after all? This is derived from d_button_proc, 
 *  but overrides the drawing routine to provide a check box.
 */
int wham_d_check_proc(int msg, DIALOG *d, int c)
{
   int x;
   int fg;
   c = 0; /* stop the compiler from complaining 'bout unused vars */

   if (msg==MSG_DRAW) {
	  fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	  text_mode(d->bg);
	  x = d->x + gui_textout(wham_guiscreen, (char *)d->dp, d->x, d->y+(d->h-(text_height(font)-gui_font_baseline))/2, fg, FALSE) + text_height(font)/2;
	  rectfill(wham_guiscreen, x+1, d->y+1, x+d->h-1, d->y+d->h-1, d->bg);
	  rect(wham_guiscreen, x, d->y, x+d->h, d->y+d->h, fg);
	  if (d->flags & D_SELECTED) {
	 line(wham_guiscreen, x, d->y, x+d->h, d->y+d->h, fg);
	 line(wham_guiscreen, x, d->y+d->h, x+d->h, d->y, fg); 
	  }
	  if (d->flags & D_GOTFOCUS)
	 dotted_rect(x+1, d->y+1, x+d->h-1, d->y+d->h-1, fg, d->bg);
	  return D_O_K;
   } 

   return wham_d_button_proc(msg, d, 0);
}



/* d_radio_proc:
 *  GUI procedure for radio buttons.
 *  Parameters: d1-button group number; d2-button style (0=circle,1=square);
 *  dp-text to appear as label to the right of the button.
 */
int wham_d_radio_proc(int msg, DIALOG *d, int c)
{
   int x, center, r, ret, fg;

   switch(msg) {

	  case MSG_DRAW:
	 fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	 text_mode(d->bg);
	 gui_textout(wham_guiscreen, (char *)d->dp, d->x+d->h+text_height(font), d->y+(d->h-(text_height(font)-gui_font_baseline))/2, fg, FALSE);

	 x = d->x;
	 r = d->h/2;
	 center = x+r;
	 rectfill(wham_guiscreen, x+1, d->y+1, x+d->h-1, d->y+d->h-1, d->bg);

	 switch (d->d2) {

		case 1:
		   rect(wham_guiscreen, x, d->y, x+d->h, d->y+d->h, fg);
		   if (d->flags & D_SELECTED)
		  rectfill(wham_guiscreen, x+r/2, d->y+r/2, x+d->h-r/2, d->y+d->h-r/2, fg);
		   break;

		default:
		   circle(wham_guiscreen, center, d->y+r, r, fg);
		   if (d->flags & D_SELECTED)
		  circlefill(wham_guiscreen, center, d->y+r, r/2, fg);
		   break;
	 }

	 if (d->flags & D_GOTFOCUS)
		dotted_rect(x+1, d->y+1, x+d->h-1, d->y+d->h-1, fg, d->bg);

	 return D_O_K;

	  case MSG_KEY:
	  case MSG_CLICK:
	 if (d->flags & D_SELECTED)
		return D_O_K;
	  break;

	  case MSG_RADIO:
	 if ((c == d->d1) && (d->flags & D_SELECTED)) {
		d->flags &= ~D_SELECTED;
		WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
	 }
	 break;
   }

   ret = d_button_proc(msg, d, 0);

   if (((msg==MSG_KEY) || (msg==MSG_CLICK)) &&
	   (d->flags & D_SELECTED) && (!(d->flags & D_CLOSE))) {
	  d->flags &= ~D_SELECTED;
	  wham_broadcast_dialog_message(MSG_RADIO, d->d1);
	  d->flags |= D_SELECTED;
   }

   return ret;
}



/* d_keyboard_proc:
 *  Invisible object for implementing keyboard shortcuts. When its key
 *  is pressed, it calls the function pointed to by dp. This should return
 *  an integer, which will be passed back to the dialog manager. The key
 *  can be specified by putting an ASCII code in the key field or by
 *  putting scancodes in d1 and d2.
 */
int wham_d_keyboard_proc(int msg, DIALOG *d, int c)
{
   int (*proc)();
   int ret = D_O_K;

   switch (msg) {

	  case MSG_START:
	 d->w = d->h = 0;
	 break;

	  case MSG_XCHAR:
	 if (((c>>8) != d->d1) && ((c>>8) != d->d2))
		break;

	 ret |= D_USED_CHAR;
	 /* fall through */

	  case MSG_KEY:
	 proc = (int (*)())d->dp;
	 ret |= (*proc)();
	 break;
   }

   return ret;
}



/* d_edit_proc:
 *  An editable text object (the dp field points to the string). When it
 *  has the input focus (obtained by clicking on it with the mouse), text
 *  can be typed into this object. The d1 field specifies the maximum
 *  number of characters that it will accept, and d2 is the text cursor 
 *  position within the string.
 */
int wham_d_edit_proc(int msg, DIALOG *d, int c)
{
   int f, l, p, w, x, fg;
   char *s;
   char buf[2];

   s = (char *)d->dp;
   l = strlen(s);
   if (d->d2 > l)
	  d->d2 = l;

   switch (msg) {

	  case MSG_START:
	 d->d2 = l;
	 break;

	  case MSG_DRAW:
	 fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	 buf[1] = 0;
	 x = 0;

	 for (p=0; p<=l; p++) {
		buf[0] = s[p] ? s[p] : ' ';
		w = text_length(font, buf);
		if (x+w > d->w)
		   break;

		f = ((p == d->d2) && (d->flags & D_GOTFOCUS));
		text_mode(f ? fg : d->bg);
		textout(wham_guiscreen, font, buf, d->x+x, d->y, f ? d->bg : fg);
		x += w;
	 }

	 if (x<d->w)
		rectfill(wham_guiscreen, d->x+x, d->y, d->x+d->w-1, d->y+text_height(font)-1, d->bg);

	 break;

	  case MSG_CLICK:
	 buf[1] = 0;
	 x = d->x;

	 for (p=0; p<l; p++) {
		buf[0] = s[p];
		x += text_length(font, buf);
		if (x > __wham_mouse_x())
		   break;
	 }

	 d->d2 = MID(0, p, l);
	 WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
	 break;

	  case MSG_WANTFOCUS:
	  case MSG_LOSTFOCUS:
	  case MSG_KEY:
	 return D_WANTFOCUS;

	  case MSG_CHAR:
	 if ((c >> 8) == KEY_LEFT) {
		if (d->d2 > 0)
		   d->d2--;
	 }
	 else if ((c >> 8) == KEY_RIGHT) {
		if (d->d2 < l)
		   d->d2++;
	 }
	 else if ((c >> 8) == KEY_HOME) {
		d->d2 = 0;
	 }
	 else if ((c >> 8) == KEY_END) {
		d->d2 = l;
	 }
	 else if ((c >> 8) == KEY_DEL) {
		if (d->d2 < l)
		   for (p=d->d2; s[p]; p++)
		  s[p] = s[p+1];
	 }
	 else if ((c >> 8) == KEY_BACKSPACE) {
		if (d->d2 > 0) {
		   d->d2--;
		   for (p=d->d2; s[p]; p++)
		  s[p] = s[p+1];
		} 
	 }
	 else {
		c &= 0xff;
		if ((c >= 32) && (c <= 126)) {
		   if (l < d->d1) {
		  while (l >= d->d2) {
			 s[l+1] = s[l];
			 l--;
		  }
		  s[d->d2] = c;
		  d->d2++;
		   }
		}
		else
		   return D_O_K;
	 }

	 /* if we changed something, better redraw... */ 
	 WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
	 return D_USED_CHAR;
   }

   return D_O_K;
}



/* draw_listbox:
 *  Helper function to draw a listbox object.
 */
static void draw_listbox(DIALOG *d)
{
   int height, listsize, i, len, bar, x, y, w;
   int fg_color, fg, bg;
   char *s;
   char store;

   (*(getfuncptr)d->dp)(-1, &listsize);
   height = (d->h-3) / text_height(font);
   bar = (listsize > height);
   w = (bar ? d->w-14 : d->w-2);
   fg_color = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;

   /* draw frame */
   rect(wham_guiscreen, d->x, d->y, d->x+d->w, d->y+d->h, fg_color);

   if (bar) {
	  vline(wham_guiscreen, d->x+d->w-12, d->y+1, d->y+d->h-1, fg_color);

	  if (d->flags & D_GOTFOCUS) {
	 dotted_rect(d->x+1, d->y+1, d->x+d->w-13, d->y+d->h-1, fg_color, d->bg);
	 dotted_rect(d->x+d->w-11, d->y+1, d->x+d->w-1, d->y+d->h-1, fg_color, d->bg);
	  }
	  else {
	 rect(wham_guiscreen, d->x+1, d->y+1, d->x+d->w-13, d->y+d->h-1, d->bg);
	 rect(wham_guiscreen, d->x+d->w-11, d->y+1, d->x+d->w-1, d->y+d->h-1, d->bg);
	  }
   }
   else {
	  if (d->flags & D_GOTFOCUS)
	 dotted_rect(d->x+1, d->y+1, d->x+d->w-1, d->y+d->h-1, fg_color, d->bg);
	  else
	 rect(wham_guiscreen, d->x+1, d->y+1, d->x+d->w-1, d->y+d->h-1, d->bg);
   }

   /* draw box contents */
   for (i=0; i<height; i++) {
	  if (d->d2+i < listsize) {
	 if (d->d2+i == d->d1) { 
		fg = d->bg;
		bg = fg_color;
	 }
	 else {
		fg = fg_color;
		bg = d->bg;
	 }
	 s = (*(getfuncptr)d->dp)(i+d->d2, NULL);
	 x = d->x + 2;
	 y = d->y + 2 + i*text_height(font);
	 text_mode(bg);
	 rectfill(wham_guiscreen, x, y, x+7, y+text_height(font)-1, bg); 
	 x += 8;
	 len = strlen(s);
	 store = 0;
	 while (text_length(font, s) >= d->w - (bar ? 22 : 10)) {
		s[len] = store;
		len--;
		store = s[len];
		s[len] = 0;
	 }
	 textout(wham_guiscreen, font, s, x, y, fg); 
	 x += text_length(font, s);
	 s[len] = store;
	 if (x <= d->x+w) 
		rectfill(wham_guiscreen, x, y, d->x+w, y+text_height(font)-1, bg);
	  }
	  else
	 rectfill(wham_guiscreen, d->x+2,  d->y+2+i*text_height(font), 
		  d->x+w, d->y+1+(i+1)*text_height(font), d->bg);
   }

   if (d->y+2+i*text_height(font) <= d->y+d->h-2)
	  rectfill(wham_guiscreen, d->x+2, d->y+2+i*text_height(font), 
					   d->x+w, d->y+d->h-2, d->bg);

   /* draw scrollbar */
   if (bar) {
	  BITMAP *pattern = create_bitmap(2, 2);
	  i = ((d->h-4) * height + listsize/2) / listsize;
	  x = d->x+d->w-10;
	  y = d->y+2;

	  pattern->line[0][1] = pattern->line[1][0] = d->bg;
	  pattern->line[0][0] = pattern->line[1][1] = fg_color;

	  if (d->d2 > 0) {
	 len = (((d->h-4) * d->d2) + listsize/2) / listsize;
	 rectfill(wham_guiscreen, x, y, x+8, y+len-1, d->bg);
	 y += len;
	  }
	  if (y+i < d->y+d->h-2) {
	 drawing_mode(DRAW_MODE_COPY_PATTERN, pattern, 0, 0);
	 rectfill(wham_guiscreen, x, y, x+8, y+i, 0);
	 solid_mode();
	 y += i;
	 rectfill(wham_guiscreen, x, y, x+8, d->y+d->h-2, d->bg);
	  }
	  else {
	 drawing_mode(DRAW_MODE_COPY_PATTERN, pattern, 0, 0);
	 rectfill(wham_guiscreen, x, y, x+8, d->y+d->h-2, 0);
	 solid_mode();
	  }

	  destroy_bitmap(pattern);
   }
}



/* scroll_listbox:
 *  Helper function to scroll through a listbox.
 */
static void scroll_listbox(DIALOG *d, int listsize)
{
   int height = (d->h-3) / text_height(font);

   if (!listsize) {
	  d->d1 = d->d2 = 0;
	  return;
   }

   /* check selected item */
   if (d->d1 < 0)
	  d->d1 = 0;
   else
	  if (d->d1 >= listsize)
	 d->d1 = listsize-1;

   /* check scroll position */
   while ((d->d2 > 0) && (d->d2 + height > listsize))
	  d->d2--;

   if (d->d2 >= d->d1) {
	  if (d->d1 < 0)
	 d->d2 = 0;
	  else
	 d->d2 = d->d1;
   }
   else {
	  while ((d->d2 + height-1) < d->d1)
	 d->d2++;
   }
}



/* _handle_listbox_click:
 *  Helper to process a click on a listbox, doing hit-testing and moving
 *  the selection.
 */
static void wham_handle_listbox_click(DIALOG *d)
{
   int listsize, height;
   int i;

   (*(getfuncptr)d->dp)(-1,&listsize);
   if (!listsize)
	  return;

   height = (d->h-3) / text_height(font);

   i = MID(0, ((__wham_mouse_y() - d->y - 2) / text_height(font)), 
			   ((d->h-3) / text_height(font) - 1));
   i += d->d2;
   if (i < d->d2)
	  i = d->d2;
   else {
	  if (i > d->d2 + height-1)
	 i = d->d2 + height-1;
	  if (i >= listsize)
	 i = listsize-1;
   }

   if (__wham_mouse_y() <= d->y)
	  i = MAX(i-1, 0);
   else if (__wham_mouse_y() >= d->y+d->h)
	  i = MIN(i+1, listsize-1);

   if (i != d->d1) {
	  d->d1 = i;
	  i = d->d2;
	  scroll_listbox(d, listsize);
	  WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
	  if (i != d->d2)
	 rest(MID(10, text_height(font)*16-d->h, 100));
   }
}



/* handle_listbox_scroll_click:
 *  Helper to process a click on a listbox scrollbar.
 */
static void handle_listbox_scroll_click(DIALOG *d)
{
   int listsize, height, i, len, x, y;

   (*(getfuncptr)d->dp)(-1, &listsize);
   height = (d->h-3) / text_height(font);

   while (mouse_b) {
	  i = ((d->h-4) * height + listsize/2) / listsize;
	  len = ((d->h) * d->d2 + listsize/2) / listsize + 2;
	  if ((__wham_mouse_y() >= d->y+len) && (__wham_mouse_y() <= d->y+len+i)) {
	 x = __wham_mouse_y() - len + 2;
	 while (mouse_b) {
		y = (listsize * (__wham_mouse_y() - x) + d->h/2) / d->h;
		if (y > listsize-height)
		   y = listsize-height;
		if (y < 0)
		   y = 0;
		if (y != d->d2) {
		   d->d2 = y;
		   WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
		}
	 }
	  }
	  else {
	 if (__wham_mouse_y() <= d->y+len)
		y = d->d2 - height;
	 else
		y = d->d2 + height;
	 if (y > listsize-height)
		y = listsize-height;
	 if (y < 0)
		y = 0;
	 if (y != d->d2) {
		d->d2 = y;
		WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
		rest(200);
	 }
	  }
   }
}



/* d_list_proc:
 *  A list box object. The dp field points to a function which it will call
 *  to obtain information about the list. This should follow the form:
 *     char *<list_func_name> (int index, int *list_size);
 *  If index is zero or positive, the function should return a pointer to
 *  the string which is to be displayed at position index in the list. If
 *  index is  negative, it should return null and list_size should be set
 *  to the number of items in the list. The list box object will allow the
 *  user to scroll through the list and to select items list by clicking
 *  on them, and if it has the input focus also by using the arrow keys. If 
 *  the D_EXIT flag is set, double clicking on a list item will cause it to 
 *  close the dialog. The index of the selected item is held in the d1 
 *  field, and d2 is used to store how far it has scrolled through the list.
 */
int wham_d_list_proc(int msg, DIALOG *d, int c)
{
   int listsize, i, bottom, height, bar;

   switch (msg) {

	  case MSG_START:
	 (*(getfuncptr)d->dp)(-1, &listsize);
	 scroll_listbox(d, listsize);
	 break;

	  case MSG_DRAW:
	 draw_listbox(d);
	 break;

	  case MSG_CLICK:
	 (*(getfuncptr)d->dp)(-1, &listsize);
	 height = (d->h-3) / text_height(font);
	 bar = (listsize > height);
	 if ((!bar) || (__wham_mouse_x() < d->x+d->w-10)) {
		while (mouse_b)
		   wham_handle_listbox_click(d);
	 }
	 else {
		handle_listbox_scroll_click(d);
	 }
	 break;

	  case MSG_DCLICK:
	 (*(getfuncptr)d->dp)(-1, &listsize);
	 height = (d->h-3) / text_height(font);
	 bar = (listsize > height);
	 if ((!bar) || (__wham_mouse_x() < d->x+d->w-10)) {
		if (d->flags & D_EXIT) {
		   if (listsize) {
		  i = d->d1;
		  WHAM_SEND_MESSAGE(d, MSG_CLICK, 0);
		  if (i == d->d1) 
			 return D_CLOSE;
		   }
		}
	 }
	 break;

	  case MSG_KEY:
	 (*(getfuncptr)d->dp)(-1, &listsize);
	 if ((listsize) && (d->flags & D_EXIT))
		return D_CLOSE;
	 break;

	  case MSG_WANTFOCUS:
	 return D_WANTFOCUS;

	  case MSG_CHAR:
	 (*(getfuncptr)d->dp)(-1,&listsize);
	 if (listsize) {
		c >>= 8;

		bottom = d->d2 + (d->h-3)/text_height(font) - 1;
		if (bottom >= listsize-1)
		   bottom = listsize-1;

		if (c == KEY_UP)
		   d->d1--;
		else if (c == KEY_DOWN)
		   d->d1++;
		else if (c == KEY_HOME)
		   d->d1 = 0;
		else if (c == KEY_END)
		   d->d1 = listsize-1;
		else if (c == KEY_PGUP) {
		   if (d->d1 > d->d2)
		  d->d1 = d->d2;
		   else
		  d->d1 -= (bottom - d->d2);
		}
		else if (c == KEY_PGDN) {
		   if (d->d1 < bottom)
		  d->d1 = bottom;
		   else
		  d->d1 += (bottom - d->d2);
		} 
		else 
		   return D_O_K;

		/* if we changed something, better redraw... */ 
		scroll_listbox(d, listsize);
		WHAM_SEND_MESSAGE(d, MSG_DRAW, 0);
		return D_USED_CHAR;
	 }
	 break;
   }

   return D_O_K;
}



/* Checking for double clicks is complicated. The user could release the
 * mouse button at almost any point, and I might miss it if I am doing some
 * other processing at the same time (eg. sending the single-click message).
 * To get around this I install a timer routine to do the checking for me,
 * so it will notice double clicks whenever they happen.
 */

static volatile int dclick_status, dclick_time;
static int dclick_install_count = 0;

#define DCLICK_START      0
#define DCLICK_RELEASE    1
#define DCLICK_AGAIN      2
#define DCLICK_NOT        3


/* dclick_check:
 *  Double click checking user timer routine.
 */
static void dclick_check(...)
{
   if (dclick_status==DCLICK_START) {              /* first click... */
	  if (!mouse_b) {
	 dclick_status = DCLICK_RELEASE;           /* aah! released first */
	 dclick_time = 0;
	 return;
	  }
   }
   else if (dclick_status==DCLICK_RELEASE) {       /* wait for second click */
	  if (mouse_b) {
	 dclick_status = DCLICK_AGAIN;             /* yes! the second click */
	 dclick_time = 0;
	 return;
	  }
   }
   else
	  return;

   /* timeout? */
   if (dclick_time++ > 10)
	  dclick_status = DCLICK_NOT;
}

static END_OF_FUNCTION(dclick_check);


void __wham_install_dclick_check()
{
   /* set up dclick checking code */
   if (dclick_install_count <= 0) {
	  LOCK_VARIABLE(dclick_status);
	  LOCK_VARIABLE(dclick_time);
	  LOCK_FUNCTION(dclick_check);
	  install_int(dclick_check, 20);
	  dclick_install_count = 1;
   }
   else
	  dclick_install_count++;

}

void __wham_remove_dclick_check()
{
   /* remove the double click handler */
   dclick_install_count--;
   if (dclick_install_count <= 0)
	  remove_int(dclick_check);

}

/* centre_dialog:
 *  Moves all the objects in a dialog so that the dialog is centered in
 *  the screen.
 */
void wham_centre_dialog(DIALOG *dialog)
{
   int min_x = INT_MAX;
   int min_y = INT_MAX;
   int max_x = INT_MIN;
   int max_y = INT_MIN;
   int xc, yc;
   int c;

   /* find the extents of the dialog */ 
   for (c=0; dialog[c].proc; c++) {
	  if (dialog[c].x < min_x)
	 min_x = dialog[c].x;

	  if (dialog[c].y < min_y)
	 min_y = dialog[c].y;

	  if (dialog[c].x + dialog[c].w > max_x)
	 max_x = dialog[c].x + dialog[c].w;

	  if (dialog[c].y + dialog[c].h > max_y)
	 max_y = dialog[c].y + dialog[c].h;
   }

   /* how much to move by? */
   xc = (wham_guiscreen->w - (max_x - min_x)) / 2 - min_x;
   yc = (wham_guiscreen->h - (max_y - min_y)) / 2 - min_y;

   /* move it */
   for (c=0; dialog[c].proc; c++) {
	  dialog[c].x += xc;
	  dialog[c].y += yc;
   }
}





/* find_dialog_focus:
 *  Searches the dialog for the object which has the input focus, returning
 *  its index, or -1 if the focus is not set. Useful after do_dialog() exits
 *  if you need to know which object was selected.
 */
int wham_find_dialog_focus(DIALOG *dialog)
{
   int c;

   for (c=0; dialog[c].proc; c++)
	  if (dialog[c].flags & D_GOTFOCUS)
	 return c;

   return -1;
}



/* dialog_message:
 *  Sends a message to all the objects in a dialog. If any of the objects
 *  return values other than D_O_K, returns the value and sets obj to the 
 *  object which produced it.
 */
int wham_dialog_message(DIALOG *dialog, int msg, int c, int *obj, Window *_the_object = NULL)
{
   int count;
   int res;
   int r;
   Window *previousWobject = NULL;

   if ((dialog != the_dialog) && !_the_object && (msg == MSG_DRAW))
		fatal("MSG_DRAW message sent to another than the current dialog\n"
			  "without specifying the container Window\n");

   if (_the_object)
		previousWobject = __wham_set_gui_container(_the_object);
   
   res = D_O_K;
   if (msg == MSG_DRAW) wham_guiobject->freeze();

   for (count=0; dialog[count].proc; count++) { 
	  if (!(dialog[count].flags & D_HIDDEN)) {
	 r = WHAM_SEND_MESSAGE(dialog+count, msg, c);
	 if (r != D_O_K) {
		res |= r;
		*obj = count;
	 }
	  }
   }
   if (msg == MSG_DRAW) 
   {
		wham_guiobject->thaw();
		wham_guiobject->update();
   }

   if (previousWobject)
		 __wham_set_gui_container(previousWobject);
   
   return res;
}



/* broadcast_dialog_message:
 *  Broadcasts a message to all the objects in the active dialog. If any of
 *  the dialog procedures return values other than D_O_K, it returns that
 *  value.
 */
int wham_broadcast_dialog_message(int msg, int c)
{
   int nowhere;

   if (the_dialog)
	  return wham_dialog_message(the_dialog, msg, c, &nowhere);
   else
	  return D_O_K;
}



/* find_mouse_object:
 *  Finds which object the mouse is on top of.
 */
static int find_mouse_object(DIALOG *d)
{
   /* finds which object the mouse is on top of */

   int mouse_object = -1;
   int c;

   for (c=0; d[c].proc; c++)
	  if ((__wham_mouse_x() >= d[c].x) && (__wham_mouse_y() >= d[c].y) &&
	  (__wham_mouse_x() < d[c].x + d[c].w) && (__wham_mouse_y() < d[c].y + d[c].h) &&
	  (!(d[c].flags & (D_HIDDEN | D_DISABLED))))
	 mouse_object = c;

   return mouse_object;
}



/* offer_focus:
 *  Offers the input focus to a particular object.
 */
static int offer_focus(DIALOG *d, int obj, int *focus_obj, int force)
{
   int res = D_O_K;

   if ((obj == *focus_obj) || 
	   ((obj >= 0) && (d[obj].flags & (D_HIDDEN | D_DISABLED))))
	  return D_O_K;

   /* check if object wants the focus */
   if (obj >= 0) {
	  res = WHAM_SEND_MESSAGE(d+obj, MSG_WANTFOCUS, 0);
	  if (res & D_WANTFOCUS)
	 res ^= D_WANTFOCUS;
	  else
	 obj = -1;
   }

   if ((obj >= 0) || (force)) {
	  /* take focus away from old object */
	  if (*focus_obj >= 0) {
	 res |= WHAM_SEND_MESSAGE(d+*focus_obj, MSG_LOSTFOCUS, 0);
	 if (res & D_WANTFOCUS) {
		if (obj < 0)
		   return D_O_K;
		else
		   res &= ~D_WANTFOCUS;
	 }
	 d[*focus_obj].flags &= ~D_GOTFOCUS;
	 res |= WHAM_SEND_MESSAGE(d+*focus_obj, MSG_DRAW, 0);
	  }

	  *focus_obj = obj;

	  /* give focus to new object */
	  if (obj >= 0) {
	 d[obj].flags |= D_GOTFOCUS;
	 res |= WHAM_SEND_MESSAGE(d+obj, MSG_GOTFOCUS, 0);
	 res |= WHAM_SEND_MESSAGE(d+obj, MSG_DRAW, 0);
	  }
   }

   return res;
}



#define MAX_OBJECTS     512

typedef struct OBJ_LIST
{
   int index;
   int diff;
} OBJ_LIST;



/* obj_list_cmp:
 *  Callback function for qsort().
 */
static int obj_list_cmp(const void *e1, const void *e2)
{
   return (((OBJ_LIST *)e1)->diff - ((OBJ_LIST *)e2)->diff);
}



/* cmp_right:
 *  Comparison function for right arrow key movement.
 */
static int cmp_right(DIALOG *d1, DIALOG *d2)
{
   int ret = (d2->x - d1->x) + ABS(d1->y - d2->y) * 8;

   if (d1->x >= d2->x)
	  ret += 0x10000;

   return ret;
}



/* cmp_left:
 *  Comparison function for left arrow key movement.
 */
static int cmp_left(DIALOG *d1, DIALOG *d2)
{
   int ret = (d1->x - d2->x) + ABS(d1->y - d2->y) * 8;

   if (d1->x <= d2->x)
	  ret += 0x10000;

   return ret;
}



/* cmp_down:
 *  Comparison function for down arrow key movement.
 */
static int cmp_down(DIALOG *d1, DIALOG *d2)
{
   int ret = (d2->y - d1->y) + ABS(d1->x - d2->x) * 8;

   if (d1->y >= d2->y)
	  ret += 0x10000;

   return ret;
}



/* cmp_up:
 *  Comparison function for up arrow key movement.
 */
static int cmp_up(DIALOG *d1, DIALOG *d2)
{
   int ret = (d1->y - d2->y) + ABS(d1->x - d2->x) * 8;

   if (d1->y <= d2->y)
	  ret += 0x10000;

   return ret;
}



/* move_focus:
 *  Handles arrow key and tab movement through a dialog, deciding which
 *  object should be given the input focus.
 */
static int move_focus(DIALOG *d, long ch, int *focus_obj)
{
   int (*cmp)(DIALOG *d1, DIALOG *d2);
   OBJ_LIST obj[MAX_OBJECTS];
   int obj_count = 0;
   int fobj, c;
   int res = D_O_K;

   /* choose a comparison function */ 
   switch (ch >> 8) {
	  case KEY_TAB:     /* same as right arrow */
	  case KEY_RIGHT:   cmp = cmp_right;  break;
	  case KEY_LEFT:    cmp = cmp_left;   break;
	  case KEY_DOWN:    cmp = cmp_down;   break;
	  case KEY_UP:      cmp = cmp_up;     break;
	  default:          return D_O_K;
   }

   /* fill temporary table */
   for (c=0; d[c].proc; c++) {
	  if ((*focus_obj < 0) || (c != *focus_obj)) {
	 obj[obj_count].index = c;
	 if (*focus_obj >= 0)
		obj[obj_count].diff = cmp(d+*focus_obj, d+c);
	 else
		obj[obj_count].diff = c;
	 obj_count++;
	 if (obj_count >= MAX_OBJECTS)
		break;
	  }
   }

   /* sort table */
   qsort(obj, obj_count, sizeof(OBJ_LIST), obj_list_cmp);

   /* find an object to give the focus to */
   fobj = *focus_obj;
   for (c=0; c<obj_count; c++) {
	  res |= offer_focus(d, obj[c].index, focus_obj, FALSE);
	  if (fobj != *focus_obj)
	 break;
   } 

   return res;
}



#define WHAM_MESSAGE(i, msg, c) {                 \
   r = WHAM_SEND_MESSAGE(dialog+i, msg, c);       \
   if (r != D_O_K) {                         \
	  res |= r;                              \
	  obj = i;                               \
   }                                         \
}
/*
 * does one step out of the while loop of do_dialog
 * assumes the arguments will  be initialised correctly elsewhere
 */

int wham_do_dialog_one_step(DIALOG *dialog,Window *w,int &obj, int &res, int &mouse_obj, long &ch,int &joy_on,int &ox,int &oy, int &focus_obj, int key_or_mouse)
{
   DIALOG *previous_dialog = the_dialog;
   BITMAP *previous_screen = wham_guiscreen;
   the_dialog = dialog;
   int 
		r,
		c;

   // select the right output bitmap
   __wham_set_gui_container(w);

	res &= ~D_USED_CHAR;
	/* need to draw it? */
	if (res & D_REDRAW) 
	{
		res ^= D_REDRAW;
		res |= wham_dialog_message(dialog, MSG_DRAW, 0, &obj);
		/* send show message to winsystem */
		//wham_guiobject->update();

	}

	/* need to give the input focus to someone? */
	if (res & D_WANTFOCUS) 
	{
	  res ^= D_WANTFOCUS;
	  res |= offer_focus(dialog, obj, &focus_obj, FALSE);
	}
	if (key_or_mouse == WHAMGUI_DO_MOUSE)
	{
		/* has mouse object changed? */
		c = find_mouse_object(dialog);
		if (c != mouse_obj) 
		{
			if (mouse_obj >= 0) 
			{
				dialog[mouse_obj].flags &= ~D_GOTMOUSE;
				WHAM_MESSAGE(mouse_obj, MSG_LOSTMOUSE, 0);
			}
			if (c >= 0) 
			{
				dialog[c].flags |= D_GOTMOUSE;
				WHAM_MESSAGE(c, MSG_GOTMOUSE, 0);
			}
			mouse_obj = c; 
	  
			/* move the input focus as well? */
			if ((gui_mouse_focus) && (mouse_obj != focus_obj))
				res |= offer_focus(dialog, mouse_obj, &focus_obj, TRUE);
	  
	  
		  }

		  /* deal with mouse button clicks */
		  if (mouse_b) 
		  {
			if (whamgui_raise_style == RAISE_CLICK)
				wham_guiobject->raise();
			res |= offer_focus(dialog, mouse_obj, &focus_obj, FALSE);
	  
			if (mouse_obj >= 0) 
			{
				dclick_time = 0;
				dclick_status = DCLICK_START;
				ox = __wham_mouse_x();
				oy = __wham_mouse_y();
	  
				/* send click message */
				WHAM_MESSAGE(mouse_obj, MSG_CLICK, 0);
	  
				if (res==D_O_K) 
				{
					do 
					{
						if ((ABS(ox-__wham_mouse_x()) > 8) || (ABS(oy-__wham_mouse_y()) > 8)) 
						{
							dclick_status = DCLICK_NOT;
							break;
						}
					} while ((dclick_status != DCLICK_AGAIN) &&
						(dclick_status != DCLICK_NOT));
	  
					/* double click! */
					if ((dclick_status==DCLICK_AGAIN) &&
						(__wham_mouse_x() >= dialog[mouse_obj].x) && 
						(__wham_mouse_y() >= dialog[mouse_obj].y) &&
						(__wham_mouse_x() <= dialog[mouse_obj].x + dialog[mouse_obj].w) &&
						(__wham_mouse_y() <= dialog[mouse_obj].y + dialog[mouse_obj].h)) 
					{
						WHAM_MESSAGE(mouse_obj, MSG_DCLICK, 0);
					}
				}
			}
			return res; // exit function
		}
	}   // end key_or_mouse == WHAMGUI_DO_MOUSE

	if (key_or_mouse == WHAMGUI_DO_KEYS)
	{  
	  /* fake joystick input by converting it to key presses */
	  if (joy_on)
			rest(20);

	  poll_joystick();

	  if (joy_on) 
	  {
		if ((!joy_left) && (!joy_right) && (!joy_up) && (!joy_down) &&
		 (!joy_b1) && (!joy_b2)) 
		{
			joy_on = FALSE;
			rest(20);
		}
		ch = 0;
	  }
	  else 
	  {
		if (joy_left) 
		{
			ch = KEY_LEFT << 8;
			joy_on = TRUE;
		}
		else if (joy_right) 
		{
			ch = KEY_RIGHT << 8;
			joy_on = TRUE;
		}
		else if (joy_up) 
		{
			ch = KEY_UP << 8;
			joy_on = TRUE;
		}
		else if (joy_down) 
		{
			ch = KEY_DOWN << 8;
			joy_on = TRUE;
		}
		else if ((joy_b1) || (joy_b2)) 
		{
			ch = (KEY_SPACE << 8) + ' ';
			joy_on = TRUE;
		}
		else
			ch = 0;
	  }

	  /* deal with keyboard input */
	  if ((ch) || (keypressed())) 
	  {
		if (!ch)
			ch = readkey();

		/* let object deal with the key? */
		if (focus_obj >= 0) 
		{
			WHAM_MESSAGE(focus_obj, MSG_CHAR, ch);


			if (res & D_USED_CHAR)
				return res; // exit function
		}

		/* keyboard shortcut? */
		if (ch & 0xff) 
		{
			for (c=0; dialog[c].proc; c++) 
			{
				if ((tolower(dialog[c].key) == tolower((ch & 0xff))) && 
				  (!(dialog[c].flags & (D_HIDDEN | D_DISABLED)))) 
				{
					WHAM_MESSAGE(c, MSG_KEY, ch);
					ch = 0;
					break;
				}
			}
			if (!ch)
				return res;
		}

		/* broadcast in case any other objects want it */
		for (c=0; dialog[c].proc; c++) 
		{
			if (!(dialog[c].flags & (D_HIDDEN | D_DISABLED))) 
			{
				WHAM_MESSAGE(c, MSG_XCHAR, ch);
		   
				if (res & D_USED_CHAR)
					return res;
			}
		}

		/* pass <CR> or <SPACE> to selected object? */
		if ((((ch & 0xff) == 10) || ((ch & 0xff) == 13) || 
		  ((ch & 0xff) == 32)) && (focus_obj >= 0)) 
		{
			WHAM_MESSAGE(focus_obj, MSG_KEY, ch);
			 return res;
		}

		/* ESC closes dialog? */
		if ((ch & 0xff) == 27) 
		{
			res |= D_CLOSE;
			obj = -1;
			 return res;
		}

		/* move focus around the dialog? */
		res |= move_focus(dialog, ch, &focus_obj);
	  }
	}  // end key_or_mouse == WHAMGUI_DO_KEYS
	the_dialog = previous_dialog;
	wham_guiscreen = previous_screen;

	  return res;
}


/* do_dialog:
 *  The basic dialog manager. The list of dialog objects should be
 *  terminated by one with a null dialog procedure. Returns the index of 
 *  the object which caused it to exit.
 */
int wham_do_dialog(DIALOG *dialog, int focus_obj, Window *the_object, int persistent = FALSE)
{

   int res = D_REDRAW;
   int obj;
   int mouse_obj;
   int c, r;
   long ch;
   int joy_on = TRUE;
   int ox, oy;
   DIALOG *previous_dialog;
   BITMAP *previous_screen;
   int 
		window_width,
		window_height;

   // store output screen
   previous_screen = wham_guiscreen;

   __wham_set_gui_container(the_object);
   the_object->show(); // unconditionally show the object
   /* set up the global  dialog pointer */
   previous_dialog = the_dialog;
   the_dialog = dialog;

   /* set up dclick checking code */
   if (dclick_install_count <= 0) {
	  LOCK_VARIABLE(dclick_status);
	  LOCK_VARIABLE(dclick_time);
	  LOCK_FUNCTION(dclick_check);
	  install_int(dclick_check, 20);
	  dclick_install_count = 1;
   }
   else
	  dclick_install_count++;

   /* initialise the dialog */
   res |= wham_dialog_message(dialog, MSG_START, 0, &obj);

   /* send show message to winsystem */
   the_object->update();
   window_width = the_object->get_geoinfo()->size_x;
   window_height = the_object->get_geoinfo()->size_y;

   mouse_obj = find_mouse_object(dialog);
   if (mouse_obj >= 0)
	  dialog[mouse_obj].flags |= D_GOTMOUSE;

   for (c=0; dialog[c].proc; c++) {
	  if (c == focus_obj)
	 dialog[c].flags |= D_GOTFOCUS;
	  else
	 dialog[c].flags &= ~D_GOTFOCUS;
   }

   /* while dialog is active */ 
   while (!(res & D_CLOSE)) 
   {
	  res &= ~D_USED_CHAR;

	  if (!persistent &&       // close dialog if clicked outside window
			( __wham_mouse_x() < 0 || __wham_mouse_x() > window_width || __wham_mouse_y() < 0 || __wham_mouse_y() > window_height)
		 )
	  {
			if (mouse_b & 1)
			{
				while (mouse_b & 1);  // wait until released
				res |= D_CLOSE;
				obj = -1;
			}
	  }


	  /* need to draw it? */
	  if (res & D_REDRAW) 
	  {
		res ^= D_REDRAW;
		res |= wham_dialog_message(dialog, MSG_DRAW, 0, &obj);
		/* send show message to winsystem */
		//the_object->update();

	  }

	  /* need to give the input focus to someone? */
	  if (res & D_WANTFOCUS) 
	  {
		res ^= D_WANTFOCUS;
		res |= offer_focus(dialog, obj, &focus_obj, FALSE);
		//the_object->update();
	  }

	  /* has mouse object changed? */
	  c = find_mouse_object(dialog);
	  if (c != mouse_obj) 
	  {
		if (mouse_obj >= 0) 
		{
			dialog[mouse_obj].flags &= ~D_GOTMOUSE;
			WHAM_MESSAGE(mouse_obj, MSG_LOSTMOUSE, 0);
		}
		if (c >= 0) 
		{
			dialog[c].flags |= D_GOTMOUSE;
			WHAM_MESSAGE(c, MSG_GOTMOUSE, 0);
		}
		mouse_obj = c; 

		/* move the input focus as well? */
		if ((gui_mouse_focus) && (mouse_obj != focus_obj))
			res |= offer_focus(dialog, mouse_obj, &focus_obj, TRUE);


	  }

	  /* deal with mouse button clicks */
	  if (mouse_b) 
	  {
		res |= offer_focus(dialog, mouse_obj, &focus_obj, FALSE);

		if (mouse_obj >= 0) 
		{
			dclick_time = 0;
			dclick_status = DCLICK_START;
			ox = __wham_mouse_x();
			oy = __wham_mouse_y();

			/* send click message */
			WHAM_MESSAGE(mouse_obj, MSG_CLICK, 0);

			if (res==D_O_K) 
			{
				do 
				{
					if ((ABS(ox-__wham_mouse_x()) > 8) || (ABS(oy-__wham_mouse_y()) > 8)) 
					{
						dclick_status = DCLICK_NOT;
						break;
					}
				} while ((dclick_status != DCLICK_AGAIN) &&
					(dclick_status != DCLICK_NOT));

				/* double click! */
				if ((dclick_status==DCLICK_AGAIN) &&
					(__wham_mouse_x() >= dialog[mouse_obj].x) && 
					(__wham_mouse_y() >= dialog[mouse_obj].y) &&
					(__wham_mouse_x() <= dialog[mouse_obj].x + dialog[mouse_obj].w) &&
					(__wham_mouse_y() <= dialog[mouse_obj].y + dialog[mouse_obj].h)) 
				{
					WHAM_MESSAGE(mouse_obj, MSG_DCLICK, 0);
				}
			}
		}
		continue;
	  }

	  /* fake joystick input by converting it to key presses */
	  if (joy_on)
			rest(20);

	  poll_joystick();

	  if (joy_on) 
	  {
		if ((!joy_left) && (!joy_right) && (!joy_up) && (!joy_down) &&
		 (!joy_b1) && (!joy_b2)) 
		{
			joy_on = FALSE;
			rest(20);
		}
		ch = 0;
	  }
	  else 
	  {
		if (joy_left) 
		{
			ch = KEY_LEFT << 8;
			joy_on = TRUE;
		}
		else if (joy_right) 
		{
			ch = KEY_RIGHT << 8;
			joy_on = TRUE;
		}
		else if (joy_up) 
		{
			ch = KEY_UP << 8;
			joy_on = TRUE;
		}
		else if (joy_down) 
		{
			ch = KEY_DOWN << 8;
			joy_on = TRUE;
		}
		else if ((joy_b1) || (joy_b2)) 
		{
			ch = (KEY_SPACE << 8) + ' ';
			joy_on = TRUE;
		}
		else
			ch = 0;
	  }

	  /* deal with keyboard input */
	  if ((ch) || (keypressed())) 
	  {
		if (!ch)
			ch = readkey();

		/* let object deal with the key? */
		if (focus_obj >= 0) 
		{
			WHAM_MESSAGE(focus_obj, MSG_CHAR, ch);


			if (res & D_USED_CHAR)
				continue;
		}

		/* keyboard shortcut? */
		if (ch & 0xff) 
		{
			for (c=0; dialog[c].proc; c++) 
			{
				if ((tolower(dialog[c].key) == tolower((ch & 0xff))) && 
				  (!(dialog[c].flags & (D_HIDDEN | D_DISABLED)))) 
				{
					WHAM_MESSAGE(c, MSG_KEY, ch);
					ch = 0;
					break;
				}
			}
			if (!ch)
				continue;
		}

		/* broadcast in case any other objects want it */
		for (c=0; dialog[c].proc; c++) 
		{
			if (!(dialog[c].flags & (D_HIDDEN | D_DISABLED))) 
			{
				WHAM_MESSAGE(c, MSG_XCHAR, ch);
		   
				if (res & D_USED_CHAR)
					continue;
			}
		}

		/* pass <CR> or <SPACE> to selected object? */
		if ((((ch & 0xff) == 10) || ((ch & 0xff) == 13) || 
		  ((ch & 0xff) == 32)) && (focus_obj >= 0)) 
		{
			WHAM_MESSAGE(focus_obj, MSG_KEY, ch);
			continue;
		}

		/* ESC closes dialog? */
		if ((ch & 0xff) == 27) 
		{
			res |= D_CLOSE;
			obj = -1;
			continue;
		}

		/* move focus around the dialog? */
		res |= move_focus(dialog, ch, &focus_obj);
	  }

	  /* send idle messages */
	  res |= wham_dialog_message(dialog, MSG_IDLE, 0, &obj);
   }

   /* send the finish messages */
   wham_dialog_message(dialog, MSG_END, 0, &obj);

   /* remove the double click handler */
   dclick_install_count--;
   if (dclick_install_count <= 0)
	  remove_int(dclick_check);

   if (mouse_obj >= 0)
	  dialog[mouse_obj].flags &= ~D_GOTMOUSE;

   the_dialog = previous_dialog;
   wham_guiscreen = previous_screen;
   return obj;
}

int wham_popup_dialog(DIALOG *dialog, int focus_obj, int persistent = FALSE)
{

   int res = D_REDRAW;
   int obj;
   int mouse_obj;
   int c, r;
   long ch;
   int joy_on = TRUE;
   int ox, oy;
   int 
		window_width = dialog->w,
		window_height = dialog->h;
   DIALOG *previous_dialog;
   BITMAP *previous_screen = wham_guiscreen;
   Window *the_object = new Window
							(
								"popup_dialog",
								dialog->w, 
								dialog->h,
								(WHAM->get_TheScreen_size_x())-dialog->w/2,
								(WHAM->get_TheScreen_size_y())-dialog->h/2
							);
   if (!the_object)
		out_of_mem("cannot create window for dialog");
   
   __wham_set_gui_container(the_object);
   the_object->show();   
   /* set up the global  dialog pointer */
   previous_dialog = the_dialog;
   the_dialog = dialog;

   /* set up dclick checking code */
   if (dclick_install_count <= 0) {
	  LOCK_VARIABLE(dclick_status);
	  LOCK_VARIABLE(dclick_time);
	  LOCK_FUNCTION(dclick_check);
	  install_int(dclick_check, 20);
	  dclick_install_count = 1;
   }
   else
	  dclick_install_count++;

   /* initialise the dialog */
   res |= wham_dialog_message(dialog, MSG_START, 0, &obj);

   /* send show message to winsystem */
   the_object->update();

   mouse_obj = find_mouse_object(dialog);
   if (mouse_obj >= 0)
	  dialog[mouse_obj].flags |= D_GOTMOUSE;

   for (c=0; dialog[c].proc; c++) {
	  if (c == focus_obj)
	 dialog[c].flags |= D_GOTFOCUS;
	  else
	 dialog[c].flags &= ~D_GOTFOCUS;
   }

   /* while dialog is active */ 
   while (!(res & D_CLOSE)) {

	  if (!persistent &&       // close dialog if clicked outside window
			( __wham_mouse_x() < 0 || __wham_mouse_x() > window_width || __wham_mouse_y() < 0 || __wham_mouse_y() > window_height)
		 )
	  {
			if (mouse_b & 1)
			{
				while (mouse_b & 1);  // wait until released
				res |= D_CLOSE;
				obj = -1;
			}
	  }

	  res &= ~D_USED_CHAR;

	  /* need to draw it? */
	  if (res & D_REDRAW) {
	 res ^= D_REDRAW;
	 res |= wham_dialog_message(dialog, MSG_DRAW, 0, &obj);
		/* send show message to winsystem */
		//the_object->update();

	  }

	  /* need to give the input focus to someone? */
	  if (res & D_WANTFOCUS) {
	 res ^= D_WANTFOCUS;
	 res |= offer_focus(dialog, obj, &focus_obj, FALSE);
		//the_object->update();
	  }

	  /* has mouse object changed? */
	  c = find_mouse_object(dialog);
	  if (c != mouse_obj) 
	  {
		if (mouse_obj >= 0) 
		{
			dialog[mouse_obj].flags &= ~D_GOTMOUSE;
			WHAM_MESSAGE(mouse_obj, MSG_LOSTMOUSE, 0);
		}
		if (c >= 0) 
		{
			dialog[c].flags |= D_GOTMOUSE;
			WHAM_MESSAGE(c, MSG_GOTMOUSE, 0);
		}
		mouse_obj = c; 

		/* move the input focus as well? */
		if ((gui_mouse_focus) && (mouse_obj != focus_obj))
			res |= offer_focus(dialog, mouse_obj, &focus_obj, TRUE);

		// send show message to window
		//the_object->update();

	  }

	  /* deal with mouse button clicks */
	  if (mouse_b) 
	  {
		res |= offer_focus(dialog, mouse_obj, &focus_obj, FALSE);

		if (mouse_obj >= 0) 
		{
			dclick_time = 0;
			dclick_status = DCLICK_START;
			ox = __wham_mouse_x();
			oy = __wham_mouse_y();

			/* send click message */
			WHAM_MESSAGE(mouse_obj, MSG_CLICK, 0);

			if (res==D_O_K) 
			{
				do 
				{
					if ((ABS(ox-__wham_mouse_x()) > 8) || (ABS(oy-__wham_mouse_y()) > 8)) 
					{
						dclick_status = DCLICK_NOT;
						break;
					}
				} while ((dclick_status != DCLICK_AGAIN) &&
					(dclick_status != DCLICK_NOT));

				/* double click! */
				if ((dclick_status==DCLICK_AGAIN) &&
					(__wham_mouse_x() >= dialog[mouse_obj].x) && 
					(__wham_mouse_y() >= dialog[mouse_obj].y) &&
					(__wham_mouse_x() <= dialog[mouse_obj].x + dialog[mouse_obj].w) &&
					(__wham_mouse_y() <= dialog[mouse_obj].y + dialog[mouse_obj].h)) 
				{
					WHAM_MESSAGE(mouse_obj, MSG_DCLICK, 0);
				}
			}
			// send show message to window
			//the_object->update();
		}
		continue;
	  }

	  /* fake joystick input by converting it to key presses */
	  if (joy_on)
	 rest(20);

	  poll_joystick();

	  if (joy_on) {
	 if ((!joy_left) && (!joy_right) && (!joy_up) && (!joy_down) &&
		 (!joy_b1) && (!joy_b2)) {
		joy_on = FALSE;
		rest(20);
	 }
	 ch = 0;
	  }
	  else {
	 if (joy_left) {
		ch = KEY_LEFT << 8;
		joy_on = TRUE;
	 }
	 else if (joy_right) {
		ch = KEY_RIGHT << 8;
		joy_on = TRUE;
	 }
	 else if (joy_up) {
		ch = KEY_UP << 8;
		joy_on = TRUE;
	 }
	 else if (joy_down) {
		ch = KEY_DOWN << 8;
		joy_on = TRUE;
	 }
	 else if ((joy_b1) || (joy_b2)) {
		ch = (KEY_SPACE << 8) + ' ';
		joy_on = TRUE;
	 }
	 else
		ch = 0;
	  }

	  /* deal with keyboard input */
	  if ((ch) || (keypressed())) {
	 if (!ch)
		ch = readkey();

	 /* let object deal with the key? */
	 if (focus_obj >= 0) {
		WHAM_MESSAGE(focus_obj, MSG_CHAR, ch);

		// send show message to window
		//the_object->update();

		if (res & D_USED_CHAR)
		   continue;
	 }

	 /* keyboard shortcut? */
	 if (ch & 0xff) {
		for (c=0; dialog[c].proc; c++) {
		   if ((tolower(dialog[c].key) == tolower((ch & 0xff))) && 
		   (!(dialog[c].flags & (D_HIDDEN | D_DISABLED)))) {
		  WHAM_MESSAGE(c, MSG_KEY, ch);
			// send show message to window
			//the_object->update();
		  ch = 0;
		  break;
		   }
		}
		if (!ch)
		   continue;
	 }

	 /* broadcast in case any other objects want it */
	 for (c=0; dialog[c].proc; c++) {
		if (!(dialog[c].flags & (D_HIDDEN | D_DISABLED))) {
		   WHAM_MESSAGE(c, MSG_XCHAR, ch);
			// send show message to window
			//the_object->update();
		   
		   if (res & D_USED_CHAR)
			continue;
		}
	 }

	 /* pass <CR> or <SPACE> to selected object? */
	 if ((((ch & 0xff) == 10) || ((ch & 0xff) == 13) || 
		  ((ch & 0xff) == 32)) && (focus_obj >= 0)) {
		WHAM_MESSAGE(focus_obj, MSG_KEY, ch);
			// send show message to window
			//the_object->update();
		continue;
	 }

	 /* ESC closes dialog? */
	 if ((ch & 0xff) == 27) {
		res |= D_CLOSE;
		obj = -1;
		continue;
	 }

	 /* move focus around the dialog? */
	 res |= move_focus(dialog, ch, &focus_obj);
	  }

	  /* send idle messages */
	  res |= wham_dialog_message(dialog, MSG_IDLE, 0, &obj);
   }

   /* send the finish messages */
   wham_dialog_message(dialog, MSG_END, 0, &obj);
   // send show message to window
   the_object->update();

   /* remove the double click handler */
   dclick_install_count--;
   if (dclick_install_count <= 0)
	  remove_int(dclick_check);

   if (mouse_obj >= 0)
	  dialog[mouse_obj].flags &= ~D_GOTMOUSE;

   the_dialog = previous_dialog;
   wham_guiscreen = previous_screen;
   delete the_object;
   return obj;
}






typedef struct MENU_INFO            /* information about a popup menu */
{
   MENU *menu;                      /* the menu itself */
   struct MENU_INFO *parent;        /* the parent menu, or NULL for root */
   int bar;                         /* set if it is a top level menu bar */
   int size;                        /* number of items in the menu */
   int sel;                         /* selected item */
   int x, y, w, h;                  /* screen position of the menu */
   int (*proc)();                   /* callback function */
   BITMAP *saved;                   /* saved what was underneath it */
} MENU_INFO;




/* get_menu_pos:
 *  Calculates the coordinates of an object within a top level menu bar.
 */
static void get_menu_pos(MENU_INFO *m, int c, int *x, int *y, int *w)
{
   int c2;

   if (m->bar) {
	  *x = 1;

	  for (c2=0; c2<c; c2++)
	 *x += gui_strlen(m->menu[c2].text) + 16;

	  *y = 1;
	  *w = gui_strlen(m->menu[c].text) + 16;
   }
   else {
	  *x = 1;
	  *y = c*(text_height(font)+4)+1;
	  *w = m->w-2;
   }
}



/* draw_menu_item:
 *  Draws an item from a popup menu onto the screen.
 */
static void draw_menu_item(MENU_INFO *m, int c)
{
   int fg, bg;
   int x, y, w;
   if (c == m->sel) {
	  fg = gui_bg_color;
	  bg = gui_fg_color;
   }
   else {
	  fg = gui_fg_color;
	  bg = gui_bg_color;
   } 

   get_menu_pos(m, c, &x, &y, &w);
   rectfill(wham_guiscreen, x, y, x+w-1, y+text_height(font)+3, bg);
   text_mode(bg);
   if (m->menu[c].text[0])
	  gui_textout(wham_guiscreen, (char *)m->menu[c].text, x+8, y+1, fg, FALSE);
   else
	  hline(wham_guiscreen, x, y+text_height(font)/2+2, x+w, fg);
}



/* draw_menu:
 *  Draws a popup menu onto the screen.
 */
static void draw_menu(MENU_INFO *m)
{
   int c;

   rect(wham_guiscreen, m->x, m->y, m->x+m->w-1, m->y+m->h-1, gui_fg_color);
   vline(wham_guiscreen, m->x+m->w, m->y+1, m->y+m->h, gui_fg_color);
   hline(wham_guiscreen, m->x+1, m->y+m->h, m->x+m->w, gui_fg_color);

   for (c=0; m->menu[c].text; c++)
	  draw_menu_item(m, c);
   wham_guiobject->update();
}



/* menu_mouse_object:
 *  Returns the index of the object the mouse is currently on top of.
 */
static int menu_mouse_object(MENU_INFO *m)
{
   int c;
   int x, y, w;

   for (c=0; c<m->size; c++) {
	  get_menu_pos(m, c, &x, &y, &w);

	  if ((__wham_mouse_x() >= x) && (__wham_mouse_x() < x+w) &&
	  (__wham_mouse_y() >= y) && (__wham_mouse_y() < y+(text_height(font)+4)))
	 return (m->menu[c].text[0]) ? c : -1;
   }

   return -1;
}



/* mouse_in_parent_menu:
 *  Recursively checks if the mouse is inside a menu or any of its parents.
 */
static int mouse_in_parent_menu(MENU_INFO *m) 
{
   int c;

   if (!m)
	  return FALSE;

   c = menu_mouse_object(m);
   if ((c >= 0) && (c != m->sel))
	  return TRUE;

   return mouse_in_parent_menu(m->parent);
}



/* fill_menu_info:
 *  Fills a menu info structure when initialising a menu.
 */
static void fill_menu_info(MENU_INFO *m, MENU *menu, MENU_INFO *parent, int bar, int x, int y)
{
   int c;

   m->menu = menu;
   m->parent = parent;
   m->bar = bar;
   m->x = x;
   m->y = y;
   m->w = 2;
   m->h = (m->bar) ? (text_height(font)+6) : 2;
   m->proc = NULL;
   m->sel = -1;

   /* calculate size of the menu */
   for (m->size=0; m->menu[m->size].text; m->size++) {
	  c = gui_strlen(m->menu[m->size].text);

	  if (m->bar) {
	 m->w += c+16;
	  }
	  else {
	 m->h += text_height(font)+4;
	 m->w = MAX(m->w, c+16);
	  }
   }
}



/* menu_key_shortcut:
 *  Returns true if c is indicated as a keyboard shortcut by a '&' character
 *  in the specified string.
 */
static int menu_key_shortcut(int c, char *s)
{
   while (*s) {
	  if (*s == '&') {
	 s++;
	 if ((*s != '&') && (tolower(*s) == tolower(c & 0xff)))
		return TRUE;
	  }
	  s++;
   }

   return FALSE;
}



/* menu_alt_key:
 *  Searches a menu for keyboard shortcuts, for the alt+letter to bring
 *  up a menu.
 */
int wham_menu_alt_key(int k, MENU *m)
{
   char *s;
   int c;

   if (k & 0xff)
	  return 0;

   k = key_ascii_table[k>>8];

   for (c=0; m[c].text; c++) 
   {
	  s = m[c].text;
	  while (*s) 
	  {
		if (*s == '&') 
		{
			s++;
			if ((*s != '&') && (tolower(*s) == tolower(k)))
				return k;
		}
		s++;
	  }
   }

   return 0;
}



/* _do_menu:
 *  The core menu control function, called by do_menu() and d_menu_proc().
 */
int wham__do_menu(MENU *menu, MENU_INFO *parent, int bar, int x, int y, int repos, int *dret)
{
   MENU_INFO m;
   MENU_INFO *i;
   int c, c2;
   int ret = -1;
   int mouse_on = mouse_b;
   int old_sel;
   int mouse_sel;
   int _x, _y;
   int redraw = TRUE;


   fill_menu_info(&m, menu, parent, bar, x, y);
   if (repos) {
	  m.x = MID(0, m.x, SCREEN_W-1 - m.w);
	  m.y = MID(0, m.y, SCREEN_H-1 - m.h);
   }
   

   Window *menu_container = new Window("menu",m.w,m.h,m.x,m.y);
   Window *old_guicontainer = __wham_set_gui_container(menu_container);
   menu_container->raise_and_lock();
   menu_container->show();

   m.sel = mouse_sel = menu_mouse_object(&m);

   if ((m.sel < 0) && (!mouse_b))
	  m.sel = 0;

		do 
		{
			old_sel = m.sel;

			c = menu_mouse_object(&m);
			if ((mouse_b) || (c != mouse_sel))
				m.sel = mouse_sel = c;

			if (mouse_b) 
			{                                  /* if button pressed */
				if ((__wham_mouse_x() < 0) || (__wham_mouse_x() > m.w) ||
					(__wham_mouse_y() < 0) || (__wham_mouse_y() > m.h)) 
				{
					if (!mouse_on)                            /* dismiss menu? */
						break;
					if (mouse_in_parent_menu(m.parent))       /* back to parent? */
						break;
				}

				if ((m.sel >= 0) && (m.menu[m.sel].child))   /* bring up child? */
					ret = m.sel;

				mouse_on = TRUE;
				clear_keybuf();
			}
			else 
			{                                          /* button not pressed */
				if (mouse_on)                                /* selected an item? */
					ret = m.sel;
				mouse_on = FALSE;

				if (keypressed()) 
				{                          /* keyboard input */
					c = readkey();

					if ((c & 0xff) == 27) 
					{
						ret = -1;
						goto getout;
					}

					switch (c >> 8) 
					{

						case KEY_LEFT:
						 if (m.parent) 
						 {
							if (m.parent->bar) 
							{
								simulate_keypress(KEY_LEFT<<8);
								simulate_keypress(KEY_DOWN<<8);
							}
							ret = -1;
							goto getout;
						 }
						/* fall through */

						case KEY_UP:
						 if ((((c >> 8) == KEY_LEFT) && (m.bar)) ||
							(((c >> 8) == KEY_UP) && (!m.bar))) 
						 {
							c = m.sel;
							do 
							{
								c--;
								if (c < 0)
									c = m.size - 1;
							} 
							while ((!(m.menu[c].text[0])) && (c != m.sel));
								m.sel = c;
						 }
						 break;

						case KEY_RIGHT:
						 if (((m.sel < 0) || (!m.menu[m.sel].child)) &&
							(m.parent) && (m.parent->bar)) 
						 {
							simulate_keypress(KEY_RIGHT<<8);
							simulate_keypress(KEY_DOWN<<8);
							ret = -1;
							goto getout;
						 }
						/* fall through */

						case KEY_DOWN:
						 if ((m.sel >= 0) && (m.menu[m.sel].child) &&
							((((c >> 8) == KEY_RIGHT) && (!m.bar)) ||
							(((c >> 8) == KEY_DOWN) && (m.bar)))) 
						 {
							ret = m.sel;
						 }
						 else if ((((c >> 8) == KEY_RIGHT) && (m.bar)) ||
							(((c >> 8) == KEY_DOWN) && (!m.bar))) 
						 {
							c = m.sel;
							do 
							{
								c++;
								if (c >= m.size)
									c = 0;
							} 
							while ((!(m.menu[c].text[0])) && (c != m.sel));
							m.sel = c;
						 }
						 break;

						case KEY_SPACE:
						case KEY_ENTER:
						 if (m.sel >= 0)
							ret = m.sel;
						 break;

						default:
						 if ((!m.parent) && ((c & 0xff) == 0))
							c = wham_menu_alt_key(c, m.menu);
						 for (c2=0; m.menu[c2].text; c2++) 
						 {
							if (menu_key_shortcut(c, m.menu[c2].text)) 
							{
								ret = m.sel = c2;
								break;
							}
						 }
						 if (m.parent) 
						 {
							i = m.parent;
							for (c2=0; i->parent; c2++)
								i = i->parent;
							c = wham_menu_alt_key(c, i->menu);
							if (c) 
							{
								while (c2-- > 0)
									simulate_keypress(27);
								simulate_keypress(c);
								ret = -1;
								goto getout;
							}
						 }
						 break;
					}
				}
			}

			if ((redraw) || (m.sel != old_sel)) 
			{           /* selection changed? */

				if (redraw) 
				{
					draw_menu(&m);
					redraw = FALSE;
				}
				else 
				{
					if (old_sel >= 0)
					draw_menu_item(&m, old_sel);

					if (m.sel >= 0)
						draw_menu_item(&m, m.sel);
					wham_guiobject->update();

				}

			}

			if (ret >= 0) 
			{                                 /* child menu? */
				if (m.menu[ret].child) 
				{
					if (m.bar) 
					{
						get_menu_pos(&m, ret, &_x, &_y, &c);
						_x += 6;
						_y += text_height(font)+7;
					}
					else 
					{
						_x = m.x+m.w*2/3;
						_y = m.y + (text_height(font)+4)*ret + text_height(font)/4+2;
					}
					c = wham__do_menu(m.menu[ret].child, &m, FALSE, _x, _y, TRUE, NULL);
					if (c < 0) 
					{
						ret = -1;
						mouse_on = FALSE;
						mouse_sel = menu_mouse_object(&m);
					}
				}
			}

			if ((m.bar) && (!mouse_b) && (!keypressed()) &&
				((__wham_mouse_x() < 0) || (__wham_mouse_x() > m.w) ||
				(__wham_mouse_y() < 0) || (__wham_mouse_y() > m.h)))
				break;
		} 
		while (ret < 0);

getout:

		if (dret)
			*dret = 0;

		/* callback function? */
		if (!m.proc)
			m.proc = m.menu[ret].proc;

		if (ret >= 0) 
		{
			if (parent)
				parent->proc = m.proc;
			else  
			{
				if (m.proc) 
				{
					c = m.proc();
					if (dret)
						*dret = c;
				}
			}
		}
		__wham_set_gui_container(old_guicontainer);
		delete menu_container;
		return ret;
}



/* do_menu:
 *  Displays and animates a popup menu at the specified screen position,
 *  returning the index of the item that was selected, or -1 if it was
 *  dismissed. If the menu crosses the edge of the screen it will be moved.
 */
int wham_do_menu(MENU *menu, int x, int y)
{
   int ret = wham__do_menu(menu, NULL, FALSE, x, y, TRUE, NULL);

   do {
   } while (mouse_b);

   return ret;
}



/* d_menu_proc:
 *  Dialog procedure for adding drop down menus to a GUI dialog. This 
 *  displays the top level menu items as a horizontal bar (eg. across the
 *  top of the screen), and pops up child menus when they are clicked.
 *  When it executes one of the menu callback routines, it passes the
 *  return value back to the dialog manager, so these can return D_O_K,
 *  D_CLOSE, D_REDRAW, etc.
 */
int wham_d_menu_proc(int msg, DIALOG *d, int c)
{ 
   MENU_INFO m;
   int ret = D_O_K;
   int x;

   switch (msg) {

	  case MSG_START:
	 fill_menu_info(&m, (MENU *)d->dp, NULL, TRUE, d->x, d->y);
	 d->w = m.w-2;
	 d->h = m.h-2;
	 break;

	  case MSG_DRAW:
	 fill_menu_info(&m, (MENU *)d->dp, NULL, TRUE, d->x, d->y);
	 draw_menu(&m);
	 break;

	  case MSG_XCHAR:
	 x = wham_menu_alt_key(c, (MENU *)d->dp);
	 if (!x)
		break;

	 ret |= D_USED_CHAR;
	 simulate_keypress(x);
	 /* fall through */

	  case MSG_GOTMOUSE:
	  case MSG_CLICK:
	 wham__do_menu((MENU *)d->dp, NULL, TRUE, d->x, d->y, FALSE, &x);
	 ret |= x;
	 do {
	 } while (mouse_b);
	 break;
   }

   return ret;
}
   

