/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/


#include <stdio.h>
#include <strings.h>
#include "../include/whamwin.h"

WhamWin::WhamWin(int _size_x,int _size_y,int _pos_x, int _pos_y, char *_name)
    : Window(_name,_size_x + 2,_size_y + 11,_pos_x,_pos_y)
{
    name = strdup(_name);
    draw_border();
    draw_bar();
    set_user_area(1,10,_size_x ,_size_y);
}

WhamWin::WhamWin(int _size_x,int _size_y,int _pos_x, int _pos_y, char *_name, WinSystem *its_manager)
    : Window(_name,_size_x + 2,_size_y + 11,_pos_x,_pos_y,its_manager)
{
    name = strdup(_name);
    draw_border();
    draw_bar();
    set_user_area(1,10,_size_x ,_size_y);
}


WhamWin::WhamWin(BITMAP *w,int _pos_x, int _pos_y, char *_name)
    : Window(_name,w,_pos_x,_pos_y)
{
    name = strdup(_name);
    draw_border();
    draw_bar();
    set_user_area(1,10,w->w - 2,w->h - 11);
}

WhamWin::WhamWin(BITMAP *w,int _pos_x, int _pos_y, char *_name, WinSystem *its_manager)
    : Window(_name,w,_pos_x,_pos_y,its_manager)
{
    name = strdup(_name);
    draw_border();
    draw_bar();
    set_user_area(1,10,w->w - 2,w->h - 11);
}


WhamWin::~WhamWin()
{
    delete name;
}

void WhamWin::draw_border()
{
    GeoInfo const *g;
    g = get_geometric_info();
    rect(bmp(),0,0,g->user_size_x - 1,g->user_size_y - 1, 15);
    // do not delete g , it's a reference
    
}

void WhamWin::draw_bar()
{
    GeoInfo const *g;
    g = get_geometric_info();

    rectfill(bmp(),1,1,g->user_size_x - 2, 9, 14);
    textout_centre(bmp(),font,(unsigned char *)name,g->user_size_x/2,1,13);
    triangle(bmp(),g->user_size_x - 3 , 8 , g->user_size_x - 11,8, g->user_size_x - 7,2,0);
    triangle(bmp(),g->user_size_x - 23 , 2, g->user_size_x - 31,2, g->user_size_x - 27,8,0);
    rectfill(bmp(),g->user_size_x - 13 , 2, g->user_size_x - 21, 8,0);
    rectfill(bmp(),g->user_size_x - 15 , 4, g->user_size_x - 19, 6,14);
}





Wobject *WhamWin::process_mouse(int *x_out = NULL, int *y_out = NULL)
{
    Wobject *win;
    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y;
    int size_was_x, size_was_y;
    int af_x,af_y;
    int flag;
    GeoInfo const *g;
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);

    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;

    if (win != this)
        return win;

    if (flag)
    {
        if (x_out <0 || y_out <0)
            set_Mouse_Cursor(WinSystem::Mouse_Cursor1);
        else
            set_Mouse_Cursor(WinSystem::Mouse_Cursor2);
    }
    else
        set_Mouse_Cursor(WinSystem::Mouse_Cursor0);

    if (mouse_b & 1)
    {
            g = win->get_geometric_info();
            if (out_y<0)
            {
                if (out_x < g->size_x - 31)
                {
                    mouse_was_x = mouse_x;
                    mouse_was_y = mouse_y;
                    pos_was_x = g->pos_x;
                    pos_was_y = g->pos_y;
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);
                    while (mouse_b & 1)
                    {

                        mouse_is_x = mouse_x;
                        mouse_is_y = mouse_y;
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);

                        win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
                    }
                }
                else if (out_x < g->size_x - 21)
                {
                    win->rescale(0,0); // will be set to minimum sizes by the window class
                    win->move(0, win->get_manager()->get_TheScreen_size_y() - win->get_manager()->Minimum_Size_y);
                }
                else if (out_x < g->size_x - 11)
                    win->rescale(g->size_x,g->size_y);
                else 
                {
                    win->rescale(win->get_manager()->get_TheScreen_size_x(),win->get_manager()->get_TheScreen_size_y());
                    win->move(0,0);
                }
            }
            else if (out_x > g->user_size_x - 10)
            {
                af_x = mouse_x;
                af_y = mouse_y;
                size_was_x = g->screen_size_x;
                size_was_y = g->screen_size_y;
                while (mouse_b & 1)
                    win->rescale(size_was_x + mouse_x - af_x,size_was_y);
            }
            else if (out_y > g->user_size_y - 10)
            {
                af_x = mouse_x;
                af_y = mouse_y;
                size_was_x = g->screen_size_x;
                size_was_y = g->screen_size_y;
                while (mouse_b & 1)
                    win->rescale(size_was_x ,size_was_y + mouse_y - af_y);
            }
            else
            {
                win->raise();
                if (g->screen_size_x <= win->get_manager()->Minimum_Size_x && g->screen_size_y <= win->get_manager()->Minimum_Size_y)
                {
                    int pos_x = g->pos_x;
                    int pos_y = g->pos_y;
                    win->rescale(g->size_x,g->size_y);
                    if (pos_x > win->get_manager()->get_TheScreen_size_x() - g->size_x)
                        pos_x =  win->get_manager()->get_TheScreen_size_x() - g->size_x;
                    if (pos_y > win->get_manager()->get_TheScreen_size_y() - g->size_y)
                        pos_y =  win->get_manager()->get_TheScreen_size_y() - g->size_y;
                    win->move(pos_x,pos_y);
                }
            }
    }
    return (Wobject *)win;
}

