/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include <stdarg.h>

#include "wtext.h"
TextWindow::TextWindow(int size_x, int size_y,int pos_x,int pos_y,  char *name, FONT *_font = NULL)
      : Wobject(name,size_x * (_font ? text_length(_font," ") : text_length(font," ") ),size_y * (_font ? text_height(_font) : text_height(font)),pos_x,pos_y,TextWindow::draw_border)
{
 	its_font = _font ? _font : font;
	cols = size_x;
   lines = size_y;
   bg_color = -1;
   fg_color = makecol(0,0,0);
   cursorx = cursory = 0;
   the_lines = (char **)malloc(sizeof(char *) * lines);
   if (! the_lines)
   	out_of_mem("could not allocate TextWindow");
      
   for (int i=0;i<lines;i++)
   {
   	the_lines[i] = (char *)malloc(sizeof(char) * (cols+1));
   	if (!the_lines[i])
	   	out_of_mem("could not allocate TextWindow");
      for (int j=0;j<cols;j++)
      	the_lines[i][j] = ' ';
      the_lines[i][cols] = 0; // ascii-z
   }
   see_through = TRUE;
   set_lock_viewport_position(TRUE);
   set_lock_viewport_size(TRUE);
   set_lock_scale(TRUE);
}

TextWindow::~TextWindow()
{
   for (int i=0;i<lines;i++)
   {
   	free(the_lines[i]);
   }
   free(the_lines);
}

void TextWindow::draw(BITMAP *onto, int f_x, int f_y,int t_x, int t_y)
{
      f_x = MAX(f_x,get_geoinfo()->pos_x);
      f_y = MAX(f_y,get_geoinfo()->pos_y);
      t_x = MIN(t_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x);
      t_y = MIN(t_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y);

      text_mode(bg_color);
      for (int i=0;i<lines;i++)
      {
      	textout(onto,its_font,the_lines[i],get_geoinfo()->pos_x,get_geoinfo()->pos_y + i * text_height(its_font), fg_color);
      }
}


int TextWindow::draw_border(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)
{
    int
        f_x,f_y,t_x,t_y;

    if (!onto) // return border dimensions to manager
    {
        switch(from_x)
        {
            case WINDOW_UPPER_BORDER:
                return 10;
            break;
            case WINDOW_LOWER_BORDER:
                return 10;
            break;
            case WINDOW_LEFT_BORDER:
                return 10;
            break;
            case WINDOW_RIGHT_BORDER:
                return 10;
            break;
        }

    }

    GeoInfo const *g = the_object->get_geoinfo();

    //upper border
    f_x = g->pos_x - 10;
    t_x = g->pos_x + g->screen_size_x + 9;
    f_y = g->pos_y - 10;      
    t_y = g->pos_y - 1;
                              
    f_x = MAX(f_x, from_x);   
    t_x = MIN(t_x, to_x);     
    f_y = MAX(f_y, from_y);   
    t_y = MIN(t_y, to_y);     
    if (f_x <= t_x && f_y <= t_y) 
        rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
    text_mode(-1);
	 textout_centre(onto,font,g->name,g->pos_x + g->screen_size_x/2,g->pos_y - 9,Title_Color);
    //lower border                         
    f_x = g->pos_x - 10;                   
    t_x = g->pos_x + g->screen_size_x + 9; 
    f_y = g->pos_y + g->screen_size_y;     
    t_y = g->pos_y + g->screen_size_y + 9; 
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
                                           
    //left border                          
    f_x = g->pos_x - 10;                   
    t_x = g->pos_x - 1;
    f_y = g->pos_y;                        
    t_y = g->pos_y + g->screen_size_y;     
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
                                           
    //right border                         
    f_x = g->pos_x + g->screen_size_x;     
    t_x = g->pos_x + g->screen_size_x + 9; 
    f_y = g->pos_y;                        
    t_y = g->pos_y + g->screen_size_y;     
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
                                           
                                           
    //END OF EXAMPLE CODE
    return 0; //returnvalue currently ignored 
}
// MOUSE FUNCTION EDIT FOR DESIRED ACTIONS
Wobject *TextWindow::process_mouse(int *x_out = NULL, int *y_out = NULL)
{
    Wobject *win; 
    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y,size_was_x,size_was_y; // helper variables
    int flag; // is the border clicked?
    GeoInfo const *g; // to store geoinfo from object in

    //get the object the mouse is currently in( if flag is set it is in a border)
    //out_x, out_y will be the coordinates relative to that object 
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);

    //store coordinates relative to this object if the user wants them
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    // check if got the right object , it could be wrong if the mouse moved very fast and moved out of the object since Wham_secretary.process_mouse  
    if (win != this) return win;

    if (flag)
    {
       if (out_x <0 || out_x <0)
          set_Mouse_Cursor(WinSystem::Mouse_Cursor1);
       else
          set_Mouse_Cursor(WinSystem::Mouse_Cursor2);
    }
    else
       set_Mouse_Cursor(WinSystem::Mouse_Cursor0);

    // mouse button 1 is currently pressed down
    if (mouse_b & 1)
    {
       g = win->get_geometric_info(); // get info on this object into g 
       mouse_was_x = mouse_x; //store current mouse position
       mouse_was_y = mouse_y;
       pos_was_x = g->pos_x; //store current object location 
       pos_was_y = g->pos_y;
       size_was_x = g->screen_size_x;//store current object size
       size_was_y = g->screen_size_y;

       //if the window has a parent manager, the parent window could be rescaled   
       //so calc the mouse position relative to the parent window
       if (win->get_manager()->get_parent())
           win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);

       if (flag) // mouse is on a border 
       {
       		raise();
//           if (out_x < 0 || out_y <0) // left or upper border -> move
//           {
               while (mouse_b & 1) // not scalable, always move
               {
                   mouse_is_x = mouse_x; // get current mouse pos
                   mouse_is_y = mouse_y;
                   // calc current mouse pos relative to parent win
                   if (win->get_manager()->get_parent())
                       win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
                   // and move the window
                   win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
               }
//           }
           /*
           else // left or lower border -> rescale
           {
               while (mouse_b & 1)
               {
                   mouse_is_x = mouse_x;// get current mosue pos
                   mouse_is_y = mouse_y;
                   // make mouse pos relative to parent window
                   if (win->get_manager()->get_parent())
                       win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
                   //rescale the object
                   win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);
               }
           }
           */
       }
       else // flag =0 -> clicked in the window
            win->raise();
    }
    if (!flag) 
        return (Wobject *)win;
    else
        return NULL;
}

void TextWindow::set_colors(int fg, int bg)
{
	fg_color = fg;
   bg_color = bg;
   if (bg_color < 0)
   	see_through = TRUE;
   else
   	see_through = FALSE;
   update();
}


void TextWindow::printf(int x, int y, char * format, ...)
{

    static char buffer[2048];
    va_list arg;
    va_start(arg,format);
    vsprintf(buffer,format,arg);
    va_end(arg);

    for (int i = x;i < cols && buffer[i - x];i++)
    {
    	the_lines[y][i] = buffer[i-x];
    }

}

void TextWindow::printf(char * format, ...)
{

    static char buffer[2048];
    va_list arg;
    va_start(arg,format);
    vsprintf(buffer,format,arg);
    va_end(arg);

    for (int i = 0;buffer[i];i++)
    {
    	if (buffer[i] != 13 && buffer[i] != 10)
      {
    		the_lines[cursory][cursorx] = buffer[i];
	      cursorx++;
   	   if (cursorx >= cols)
      	{
      		cursory++;
	         cursorx = 0;
   	   }
      }
      else if  (buffer[i] == 10)
      {
      	cursory++;
         cursorx = 0;
      }
    	if (cursory >=lines)
     		scrollup();
      
    }
    update();
}

void TextWindow::scrollup(int _lines = 1)
{
   if (_lines >= lines)
   {
	   clear();
      return;
   }
   char *firstline;
   for (int i =0;i<_lines;i++)
   {
		firstline = the_lines[0];
      
      for (int j = 0;j< lines - 1;j++)
      	the_lines[j] = the_lines[j + 1];
         
		for (int j = 0; j<cols;j++)
      	firstline[j] = ' ';
         
      the_lines[lines - 1] = firstline;
      
   }
   cursory -= _lines;
   update();
}

void TextWindow::clear()
{
   for (int i=0;i<lines;i++)
   {
      for (int j=0;j<cols;j++)
      	the_lines[i][j] = ' ';
   }
   cursorx = 0;
   cursory = 0;
   update();
}

void TextWindow::gotoxy(int x, int y)
{
	x = MID(0,x,cols - 1);
	y = MID(0,y,lines - 1);
   cursorx = x;
   cursory = y;
}

int TextWindow::Border_Color = 255;
int TextWindow::Title_Color = 0;
