# makefile for gnu make for the WHAM windowing library
# requires DJGPPv2.01+ and Allegro3.0+

#remove gui.o and guiproc.o when using an allegro version higher than 3.0
OBJS = obj/winsys.o obj/wobject.o obj/helpers.o obj/whamwin.o obj/bordered.o obj/window.o  \
	 obj/secretar.o obj/whamgui.o obj/guicon.o obj/billwin.o obj/frame.o obj/whaminit.o \
	obj/framewin.o obj/panel.o obj/pager.o obj/boxlist.o obj/scrolbar.o obj/sunglass.o \
	obj/oclock.o obj/geoinfo.o obj/scrsav.o obj/alguicon.o  obj/wtext.o\
	obj/gui.o obj/guiproc.o

CFLAGS = -O3 -Wall -ffast-math -fomit-frame-pointer -finline-functions \
	 -I./include  -g -Werror  

#remove thel line below when using a higher evrsion of Allegro than 3.0
ALLEGPATH = g:/

#Debugging flags
#uncomment this if you want to be able to see the hidingsidplaying of windows (only works in unbuffered mode)
#CFLAGS += -DSLOW
#shows internal function trace of the library
#CFLAGS += -DTRACE
#to debug the boxlist algorithm (try this one in unbuffered mode :)
#CFLAGS += -DDEBUGBOXLIST
#to debug message system
#CFLAGS += -DDEBUGMESSAGESYS


#headers used by all sourcefiles
BASEHEADERS = include/winsys.h include/wobject.h include/geoinfo.h \
		 include/helpers.h include/boxlist.h include/topinc.h\
		include/wversion.h

# headers used by sourcefiles derived from Window
BASEHEADERSW = include/winsys.h include/wobject.h include/geoinfo.h \
		 include/helpers.h include/window.h include/topinc.h \
		include/wversion.h

#headers used by sourcefiles derived from BillWin
BASEHEADERSB = include/winsys.h include/wobject.h include/geoinfo.h \
		 include/helpers.h include/billwin.h include/topinc.h \
		include/wversion.h


lib/libwham.a : $(OBJS)
	del lib\libwham.a
	ar -rs ./lib/libwham.a $(OBJS)   

obj/pager.o: src/pager.cc include/pager.h $(BASEHEADERS)
	gcc -c -o obj/pager.o $(CFLAGS) src/pager.cc

obj/panel.o: src/panel.cc include/panel.h $(BASEHEADERSW)
	gcc -c -o obj/panel.o $(CFLAGS) src/panel.cc

obj/framewin.o: src/framewin.cc include/framewin.h $(BASEHEADERSW)
	gcc -c -o obj/framewin.o $(CFLAGS) src/framewin.cc

obj/whaminit.o: src/whaminit.cc include/whaminit.h include/billwin.h include/bordered.h \
	include/window.h include/whamwin.h include/guicon.h $(BASEHEADERS)
	gcc -c -o obj/whaminit.o $(CFLAGS) src/whaminit.cc

obj/billwin.o: src/billwin.cc include/billwin.h $(BASEHEADERS)
	gcc -c -o obj/billwin.o $(CFLAGS) src/billwin.cc

obj/guicon.o: src/guicon.cc include/guicon.h include/whamgui.h $(BASEHEADERSW)
	gcc -c -o obj/guicon.o $(CFLAGS) src/guicon.cc

obj/alguicon.o: src/alguicon.cc include/alguicon.h $(BASEHEADERSW)
	gcc -c -o obj/alguicon.o $(CFLAGS) src/alguicon.cc

obj/whamgui.o: src/whamgui.cc include/whamgui.h $(BASEHEADERS)
	gcc -o obj/whamgui.o $(CFLAGS) -c src/whamgui.cc

obj/secretar.o: src/secretar.cc include/secretar.h include/scrsav.h $(BASEHEADERS)
	gcc -o obj/secretar.o $(CFLAGS) -c src/secretar.cc

obj/frame.o: src/frame.cc include/frame.h $(BASEHEADERS)
	gcc -o obj/frame.o $(CFLAGS) -c src/frame.cc

obj/window.o: src/window.cc include/window.h $(BASEHEADERS)
	gcc -o obj/window.o $(CFLAGS) -c src/window.cc

obj/winsys.o : src/winsys.cc $(BASEHEADERS)
	gcc -o obj/winsys.o $(CFLAGS) -c src/winsys.cc

obj/wobject.o : src/wobject.cc $(BASEHEADERS)
	gcc -o obj/wobject.o $(CFLAGS) -c src/wobject.cc

obj/helpers.o : src/helpers.cc $(BASEHEADERS)
	gcc -o obj/helpers.o $(CFLAGS) -c src/helpers.cc

obj/whamwin.o : src/whamwin.cc include/whamwin.h $(BASEHEADERSW)
	gcc -o obj/whamwin.o $(CFLAGS) -c src/whamwin.cc

obj/bordered.o: src/bordered.cc include/bordered.h $(BASEHEADERSW)
	gcc -o obj/bordered.o $(CFLAGS) -c src/bordered.cc

obj/boxlist.o : src/boxlist.cc $(BASEHEADERS)
	gcc -o obj/boxlist.o $(CFLAGS) -c src/boxlist.cc

obj/scrolbar.o : src/scrolbar.cc include/scrolbar.h $(BASEHEADERS)
	gcc -o obj/scrolbar.o $(CFLAGS) -c src/scrolbar.cc

obj/sunglass.o : src/sunglass.cc include/sunglass.h $(BASEHEADERS)
	gcc -o obj/sunglass.o $(CFLAGS) -c src/sunglass.cc

obj/oclock.o : src/oclock.cc include/oclock.h $(BASEHEADERS)
	gcc -o obj/oclock.o $(CFLAGS) -c src/oclock.cc

obj/geoinfo.o: src/geoinfo.cc include/geoinfo.h $(BASEHEADERS)
	gcc -o obj/geoinfo.o $(CFLAGS) -c src/geoinfo.cc

obj/scrsav.o: src/scrsav.cc include/screensv.h include/scrsav.h include/wversion.h
	gcc -o obj/scrsav.o $(CFLAGS) -c src/scrsav.cc

obj/wtext.o: src/wtext.cc include/wtext.h $(BASEHEADERS)
	gcc -c -o obj/wtext.o $(CFLAGS) src/wtext.cc


# remove these targets when using an allegro version higher than 3.0
obj/gui.o: src/gui.c 
	gcc -o obj/gui.o $(CFLAGS) -c src/gui.c -I$(ALLEGPATH)/src -I$(ALLEGPATH)/src/djgpp

obj/guiproc.o: src/guiproc.c 
	gcc -o obj/guiproc.o $(CFLAGS) -c src/guiproc.c -I$(ALLEGPATH)/src -I$(ALLEGPATH)/src/djgpp


   
